# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from urllib.parse import urlparse

import storey
from mergedeep import merge
from storey import V3ioDriver

import mlrun
from mlrun.datastore.base import DataStore
from mlrun.datastore.datastore_profile import (
    DatastoreProfileKafkaStream,
    DatastoreProfileKafkaTarget,
    DatastoreProfileTDEngine,
    datastore_profile_read,
)

from ..platforms.iguazio import parse_path
from .utils import (
    parse_kafka_url,
)

"""
Storey targets expect storage_options, which may contain credentials.
To avoid passing it openly within the graph, we use wrapper classes.
"""


def get_url_and_storage_options(path, external_storage_options=None):
    store, resolved_store_path, url = mlrun.store_manager.get_or_create_store(path)
    storage_options = store.get_storage_options()
    if storage_options and external_storage_options:
        # merge external storage options with the store's storage options. storage_options takes precedence
        storage_options = merge(external_storage_options, storage_options)
    else:
        storage_options = storage_options or external_storage_options
    return url, DataStore._sanitize_options(storage_options)


class TDEngineStoreyTarget(storey.TDEngineTarget):
    def __init__(self, *args, url: str, **kwargs):
        if url.startswith("ds://"):
            datastore_profile = datastore_profile_read(url)
            if not isinstance(datastore_profile, DatastoreProfileTDEngine):
                raise ValueError(
                    f"Unexpected datastore profile type:{datastore_profile.type}."
                    "Only DatastoreProfileTDEngine is supported"
                )
            url = datastore_profile.dsn()
        super().__init__(*args, url=url, **kwargs)


class StoreyTargetUtils:
    @staticmethod
    def process_args_and_kwargs(args, kwargs):
        args = list(args)
        path = args[0] if args else kwargs.get("path")
        external_storage_options = kwargs.get("storage_options")

        url, storage_options = get_url_and_storage_options(
            path, external_storage_options
        )

        if storage_options:
            kwargs["storage_options"] = storage_options
        if args:
            args[0] = url
        if "path" in kwargs:
            kwargs["path"] = url
        return args, kwargs


class ParquetStoreyTarget(storey.ParquetTarget):
    def __init__(self, *args, **kwargs):
        alt_key_name = kwargs.pop("alternative_v3io_access_key", None)
        args, kwargs = StoreyTargetUtils.process_args_and_kwargs(args, kwargs)
        storage_options = kwargs.get("storage_options", {})
        if storage_options and storage_options.get("v3io_access_key") and alt_key_name:
            if alt_key := mlrun.get_secret_or_env(alt_key_name):
                storage_options["v3io_access_key"] = alt_key
        super().__init__(*args, **kwargs)


class CSVStoreyTarget(storey.CSVTarget):
    def __init__(self, *args, **kwargs):
        args, kwargs = StoreyTargetUtils.process_args_and_kwargs(args, kwargs)
        super().__init__(*args, **kwargs)


class StreamStoreyTarget(storey.StreamTarget):
    def __init__(self, *args, **kwargs):
        args = list(args)

        uri = args[0] if args else kwargs.get("stream_path")

        if not uri:
            raise mlrun.errors.MLRunInvalidArgumentError("StreamTarget requires a path")

        _, storage_options = get_url_and_storage_options(uri)
        _, path = parse_path(uri)

        access_key = storage_options.get("v3io_access_key")

        if alt_key_name := kwargs.pop("alternative_v3io_access_key", None):
            if alt_key := mlrun.get_secret_or_env(alt_key_name):
                access_key = alt_key

        storage = V3ioDriver(access_key=access_key)

        if storage_options:
            kwargs["storage"] = storage
        if args:
            args[0] = path
        if "stream_path" in kwargs:
            kwargs["stream_path"] = path

        super().__init__(*args, **kwargs)


class KafkaStoreyTarget(storey.KafkaTarget):
    def __init__(self, *args, **kwargs):
        kwargs.pop("alternative_v3io_access_key", None)
        path = kwargs.pop("path")
        attributes = kwargs.pop("attributes", {})
        if path and path.startswith("ds://"):
            datastore_profile = datastore_profile_read(path)
            if not isinstance(
                datastore_profile,
                (DatastoreProfileKafkaStream, DatastoreProfileKafkaTarget),
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Unsupported datastore profile type: {type(datastore_profile)}"
                )

            attributes = merge(attributes, datastore_profile.attributes())
            brokers = attributes.pop("brokers", None)
            # Override the topic with the one in the url (if any)
            parsed = urlparse(path)
            topic = (
                parsed.path.strip("/") if parsed.path else datastore_profile.get_topic()
            )
        else:
            brokers = attributes.pop("brokers", None)
            topic, brokers = parse_kafka_url(path, brokers)

        if not topic:
            raise mlrun.errors.MLRunInvalidArgumentError("KafkaTarget requires a topic")
        kwargs["brokers"] = brokers
        kwargs["topic"] = topic

        attributes = mlrun.datastore.utils.KafkaParameters(attributes).producer()

        super().__init__(*args, **kwargs, producer_options=attributes)


class NoSqlStoreyTarget(storey.NoSqlTarget):
    pass


class RedisNoSqlStoreyTarget(storey.NoSqlTarget):
    def __init__(self, *args, **kwargs):
        path = kwargs.pop("path")
        endpoint, uri = mlrun.datastore.targets.RedisNoSqlTarget.get_server_endpoint(
            path
        )
        kwargs["path"] = endpoint + "/" + uri
        super().__init__(*args, **kwargs)


class TSDBStoreyTarget(storey.TSDBTarget):
    pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util as imputil
import json
import os
import pathlib
import socket
import sys
import tempfile
import time
import typing
import uuid
from base64 import b64decode
from copy import deepcopy
from os import environ, makedirs, path
from pathlib import Path
from typing import Optional, Union

import nuclio
import yaml

import mlrun.common.constants as mlrun_constants
import mlrun.common.formatters
import mlrun.common.schemas
import mlrun.errors
import mlrun.utils.helpers
import mlrun_pipelines.utils
from mlrun.datastore.model_provider.model_provider import ModelProvider
from mlrun_pipelines.common.models import RunStatuses
from mlrun_pipelines.common.ops import format_summary_from_kfp_run, show_kfp_run

from .common.helpers import parse_versioned_object_uri
from .config import config as mlconf
from .datastore import store_manager
from .errors import MLRunInvalidArgumentError, MLRunTimeoutError
from .execution import MLClientCtx
from .model import RunObject, RunTemplate
from .runtimes import (
    DaskCluster,
    HandlerRuntime,
    KubejobRuntime,
    LocalRuntime,
    MpiRuntimeV1,
    RemoteRuntime,
    RemoteSparkRuntime,
    RuntimeKinds,
    ServingRuntime,
    Spark3Runtime,
    get_runtime_class,
)
from .runtimes.databricks_job.databricks_runtime import DatabricksRuntime
from .runtimes.funcdoc import update_function_entry_points
from .runtimes.nuclio.application import ApplicationRuntime
from .runtimes.utils import add_code_metadata, global_context
from .utils import (
    RunKeys,
    create_ipython_display,
    extend_hub_uri_if_needed,
    get_in,
    logger,
    retry_until_successful,
    update_in,
)

if typing.TYPE_CHECKING:
    from mlrun.datastore import DataItem


def function_to_module(code="", workdir=None, secrets=None, silent=False):
    """Load code, notebook or mlrun function as .py module
    this function can import a local/remote py file or notebook
    or load an mlrun function object as a module, you can use this
    from your code, notebook, or another function (for common libs)

    Note: the function may have package requirements which must be satisfied

    example::

        mod = mlrun.function_to_module("./examples/training.py")
        task = mlrun.new_task(inputs={"infile.txt": "../examples/infile.txt"})
        context = mlrun.get_or_create_ctx("myfunc", spec=task)
        mod.my_job(context, p1=1, p2="x")
        print(context.to_yaml())

        fn = mlrun.import_function("hub://open-archive")
        mod = mlrun.function_to_module(fn)
        data = mlrun.run.get_dataitem(
            "https://fpsignals-public.s3.amazonaws.com/catsndogs.tar.gz"
        )
        context = mlrun.get_or_create_ctx("myfunc")
        mod.open_archive(context, archive_url=data)
        print(context.to_yaml())

    :param code:    path/url to function (.py or .ipynb or .yaml)
                    OR function object
    :param workdir: code workdir
    :param secrets: secrets needed to access the URL (e.g.s3, v3io, ..)
    :param silent:  do not raise on errors

    :returns: python module
    """
    command, _ = load_func_code(code, workdir, secrets=secrets)
    if not command:
        if silent:
            return None
        raise ValueError("nothing to run, specify command or function")

    command = os.path.join(workdir or "", command)

    source_file_path_object, working_dir_path_object = (
        mlrun.utils.helpers.get_source_and_working_dir_paths(command)
    )
    if source_file_path_object.is_relative_to(working_dir_path_object):
        mod_name = mlrun.utils.helpers.get_relative_module_name_from_path(
            source_file_path_object, working_dir_path_object
        )
    elif source_file_path_object.is_relative_to(
        pathlib.Path(tempfile.gettempdir()).resolve()
    ):
        mod_name = Path(command).stem
    else:
        raise mlrun.errors.MLRunRuntimeError(
            f"Cannot run source file '{command}': it must be located either under the current working "
            f"directory ('{working_dir_path_object}') or the system temporary directory ('{tempfile.gettempdir()}'). "
            f"This is required when running with local=True."
        )

    spec = imputil.spec_from_file_location(mod_name, command)
    if spec is None:
        raise OSError(f"cannot import from {command!r}")
    mod = imputil.module_from_spec(spec)
    # add to system modules, which can be necessary when running in a MockServer (ML-10937)
    sys.modules[mod_name] = mod
    spec.loader.exec_module(mod)

    return mod


def load_func_code(command="", workdir=None, secrets=None, name="name"):
    is_obj = hasattr(command, "to_dict")
    suffix = "" if is_obj else Path(command).suffix
    runtime = None
    if is_obj or suffix == ".yaml":
        is_remote = False
        if is_obj:
            runtime = command
        else:
            is_remote = "://" in command
            data = get_object(command, secrets)
            runtime = yaml.safe_load(data)
            runtime = new_function(runtime=runtime)

        command = runtime.spec.command or ""
        code = runtime.spec.build.functionSourceCode
        origin_filename = runtime.spec.build.origin_filename
        kind = runtime.kind or ""
        if kind in RuntimeKinds.nuclio_runtimes():
            code = get_in(runtime.spec.base_spec, "spec.build.functionSourceCode", code)
        if code:
            if (
                origin_filename
                and origin_filename.endswith(".py")
                and path.isfile(origin_filename)
            ):
                command = origin_filename
            else:
                suffix = ".py"
                if origin_filename:
                    suffix = f"-{pathlib.Path(origin_filename).stem}.py"
                with tempfile.NamedTemporaryFile(
                    suffix=suffix, mode="w", delete=False
                ) as temp_file:
                    code = b64decode(code).decode("utf-8")
                    command = temp_file.name
                    temp_file.write(code)

        elif command and not is_remote:
            file_path = path.join(workdir or "", command)
            if not path.isfile(file_path):
                raise OSError(f"command file {file_path} not found")

        else:
            logger.warn("run command, file or code were not specified")

    elif command == "":
        pass

    elif suffix == ".ipynb":
        temp_file = tempfile.NamedTemporaryFile(suffix=".py", delete=False)
        code_to_function(
            name, filename=command, kind="local", code_output=temp_file.name
        )
        command = temp_file.name

    elif suffix == ".py":
        if "://" in command:
            temp_file = tempfile.NamedTemporaryFile(suffix=".py", delete=False)
            download_object(command, temp_file.name, secrets)
            command = temp_file.name

    else:
        raise ValueError(f"unsupported suffix: {suffix}")

    return command, runtime


def get_or_create_ctx(
    name: str,
    event=None,
    spec: Optional[dict] = None,
    with_env: bool = True,
    rundb: Union[str, "mlrun.db.RunDBInterface"] = "",
    project: str = "",
    upload_artifacts: bool = False,
) -> MLClientCtx:
    """
    Called from within the user program to obtain a run context.

    The run context is an interface for receiving parameters, data and logging
    run results, the run context is read from the event, spec, or environment
    (in that order), user can also work without a context (local defaults mode).

    all results are automatically stored in the "rundb" or artifact store,
    the path to the rundb can be specified in the call or obtained from env.

    :param name:     run name (will be overridden by context)
    :param event:    function (nuclio Event object)
    :param spec:     dictionary holding run spec
    :param with_env: look for context in environment vars, default True
    :param rundb:    path/url to the metadata and artifact database
    :param project:  project to initiate the context in (by default `mlrun.mlconf.active_project`).
                              If not set, an active project must exist.
    :param upload_artifacts:  when using local context (not as part of a job/run), upload artifacts to the
                              system default artifact path location
    :return: execution context

    Examples::

        # load MLRUN runtime context (will be set by the runtime framework e.g. KubeFlow)
        context = get_or_create_ctx("train")

        # get parameters from the runtime context (or use defaults)
        p1 = context.get_param("p1", 1)
        p2 = context.get_param("p2", "a-string")

        # access input metadata, values, files, and secrets (passwords)
        print(f"Run: {context.name} (uid={context.uid})")
        print(f"Params: p1={p1}, p2={p2}")
        print(f'accesskey = {context.get_secret("ACCESS_KEY")}')
        input_str = context.get_input("infile.txt").get()
        print(f"file: {input_str}")

        # RUN some useful code e.g. ML training, data prep, etc.

        # log scalar result values (job result metrics)
        context.log_result("accuracy", p1 * 2)
        context.log_result("loss", p1 * 3)
        context.set_label("framework", "sklearn")

        # log various types of artifacts (file, web page, table), will be versioned and visible in the UI
        context.log_artifact(
            "model.txt", body=b"abc is 123", labels={"framework": "xgboost"}
        )
        context.log_artifact("results.html", body=b"<b> Some HTML <b>", viewer="web-app")

    """
    if global_context.get() and not spec and not event:
        return global_context.get()

    newspec = {}
    config = environ.get("MLRUN_EXEC_CONFIG")
    if event:
        newspec = event.body

    elif spec:
        newspec = deepcopy(spec)

    elif with_env and config:
        newspec = config

    if isinstance(newspec, (RunObject, RunTemplate)):
        newspec = newspec.to_dict()

    if newspec and not isinstance(newspec, dict):
        newspec = json.loads(newspec)

    if (
        not newspec.get("metadata", {}).get("project")
        and not project
        and not mlconf.active_project
    ):
        raise mlrun.errors.MLRunMissingProjectError(
            """No active project found. Make sure to set an active project using: mlrun.get_or_create_project()
            You can verify the active project with: mlrun.mlconf.active_project"""
        )

    if not newspec:
        newspec = {}
        if upload_artifacts:
            artifact_path = mlrun.utils.helpers.template_artifact_path(
                mlconf.artifact_path, project or mlconf.active_project
            )
            update_in(newspec, ["spec", RunKeys.output_path], artifact_path)

    newspec.setdefault("metadata", {})
    update_in(newspec, "metadata.name", name, replace=False)
    autocommit = False
    tmp = environ.get("MLRUN_META_TMPFILE")
    out = rundb or mlconf.dbpath or environ.get("MLRUN_DBPATH")
    if out:
        autocommit = True
        logger.info(f"Logging run results to: {out}")

    newspec["metadata"]["project"] = (
        newspec["metadata"].get("project") or project or mlconf.active_project
    )

    newspec["metadata"].setdefault("labels", {})

    # This function can also be called as a local run if it is not called within a function.
    # It will create a local run, and the run kind must be local by default.
    newspec["metadata"]["labels"].setdefault(
        mlrun_constants.MLRunInternalLabels.kind, RuntimeKinds.local
    )

    ctx = MLClientCtx.from_dict(
        newspec, rundb=out, autocommit=autocommit, tmp=tmp, host=socket.gethostname()
    )
    global_context.set(ctx)
    return ctx


def import_function(url="", secrets=None, db="", project=None, new_name=None):
    """Create function object from DB or local/remote YAML file

    Functions can be imported from function repositories (mlrun Function Hub (formerly Marketplace) or local db),
    or be read from a remote URL (http(s), s3, git, v3io, ..) containing the function YAML

    special URLs::

        function hub:       hub://[{source}/]{name}[:{tag}]
        local mlrun db:     db://{project-name}/{name}[:{tag}]

    examples::

        function = mlrun.import_function("hub://auto-trainer")
        function = mlrun.import_function("./func.yaml")
        function = mlrun.import_function(
            "https://raw.githubusercontent.com/org/repo/func.yaml"
        )

    :param url: path/url to Function Hub, db or function YAML file
    :param secrets: optional, credentials dict for DB or URL (s3, v3io, ...)
    :param db: optional, mlrun api/db path
    :param project: optional, target project for the function
    :param new_name: optional, override the imported function name

    :returns: function object
    """
    is_hub_uri = False
    if url.startswith("db://"):
        url = url[5:]
        _project, name, tag, hash_key = parse_versioned_object_uri(url)
        db = mlrun.db.get_run_db(db or mlrun.db.get_or_set_dburl(), secrets=secrets)
        runtime = db.get_function(name, _project, tag, hash_key)
        if not runtime:
            raise KeyError(f"function {name}:{tag} not found in the DB")
    else:
        url, is_hub_uri = extend_hub_uri_if_needed(url)
        runtime = import_function_to_dict(url, secrets)
    function = new_function(runtime=runtime)
    project = project or mlrun.mlconf.active_project
    # When we're importing from the hub we want to assign to a target project, otherwise any store on it will
    # simply default to the active project
    if project and is_hub_uri:
        function.metadata.project = project
    if new_name:
        function.metadata.name = mlrun.utils.helpers.normalize_name(new_name)
    return function


def import_function_to_dict(
    url: str,
    secrets: Optional[dict] = None,
) -> dict:
    """Load function spec from local/remote YAML file"""
    obj = get_object(url, secrets)
    runtime = yaml.safe_load(obj)
    remote = "://" in url

    code = get_in(runtime, "spec.build.functionSourceCode")
    update_in(runtime, "metadata.build.code_origin", url)
    cmd = code_file = get_in(runtime, "spec.command", "")
    # use kind = "job" by default if not specified
    runtime.setdefault("kind", "job")
    if " " in cmd:
        code_file = cmd[: cmd.find(" ")]
    if runtime["kind"] in ["", "local"]:
        if code:
            with tempfile.NamedTemporaryFile(
                suffix=".py", mode="w", delete=False
            ) as temp_file:
                code = b64decode(code).decode("utf-8")
                update_in(runtime, "spec.command", temp_file.name)
                temp_file.write(code)
        elif remote and cmd:
            if cmd.startswith("/"):
                raise ValueError("exec path (spec.command) must be relative")
            url = url[: url.rfind("/") + 1] + code_file
            code = get_object(url, secrets)
            code_file = _ensure_path_confined_to_base_dir(
                base_directory=".",
                relative_path=code_file,
                error_message_on_escape="Path traversal detected in spec.command",
            )
            dir = path.dirname(code_file)
            if dir:
                makedirs(dir, exist_ok=True)
            with open(code_file, "wb") as fp:
                fp.write(code)
        elif cmd:
            slash_index = url.rfind("/")
            if slash_index < 0:
                raise ValueError(f"no file in exec path (spec.command={code_file})")
            base_dir = os.path.normpath(url[: slash_index + 1])

            # Validate and resolve the candidate path before checking existence
            candidate_path = _ensure_path_confined_to_base_dir(
                base_directory=base_dir,
                relative_path=code_file,
                error_message_on_escape=(
                    f"exec file spec.command={code_file} is outside of allowed directory"
                ),
            )

            # Only now it's safe to check file existence
            if not path.isfile(candidate_path):
                raise ValueError(f"no file in exec path (spec.command={code_file})")

            # Check that the path is absolute
            if not os.path.isabs(code_file):
                raise ValueError(
                    f"exec file spec.command={code_file} is relative, it must be absolute. Change working dir"
                )
        else:
            raise ValueError("command or code not specified in function spec")

    return runtime


def new_function(
    name: Optional[str] = "",
    project: Optional[str] = "",
    tag: Optional[str] = "",
    kind: Optional[str] = "",
    command: Optional[str] = "",
    image: Optional[str] = "",
    args: Optional[list] = None,
    runtime: Optional[Union[mlrun.runtimes.BaseRuntime, dict]] = None,
    mode: Optional[str] = None,
    handler: Optional[str] = None,
    source: Optional[str] = None,
    requirements: Optional[list[str]] = None,
    kfp: Optional[bool] = None,
    requirements_file: str = "",
):
    """Create a new ML function from base properties

    Example::

           # define a container based function (the `training.py` must exist in the container workdir)
           f = new_function(
               command="training.py -x {x}", image="myrepo/image:latest", kind="job"
           )
           f.run(params={"x": 5})

           # define a container based function which reads its source from a git archive
           f = new_function(
               command="training.py -x {x}",
               image="myrepo/image:latest",
               kind="job",
               source="git://github.com/mlrun/something.git",
           )
           f.run(params={"x": 5})

           # define a local handler function (execute a local function handler)
           f = new_function().run(task, handler=myfunction)

    :param name:     function name
    :param project:  function project (none for the active project)
    :param tag:      function version tag (none for 'latest')

    :param kind:     runtime type (local, job, nuclio, spark, mpijob, dask, ..)
    :param command:  command/url + args (e.g.: training.py --verbose)
    :param image:    container image (start with '.' for default registry)
    :param args:     command line arguments (override the ones in command)
    :param runtime:  runtime (job, nuclio, spark, dask ..) object/dict
                     store runtime specific details and preferences
    :param mode:     runtime mode:
        * pass - will run the command as is in the container (not wrapped by mlrun), the command can use
        params substitutions like {xparam} and will be replaced with the value of the xparam param
        if a command is not specified, then image entrypoint shall be used.
    :param handler:  The default function handler to call for the job or nuclio function, in batch functions
         (job, mpijob, ..) the handler can also be specified in the `.run()` command, when not specified
         the entire file will be executed (as main).
         for nuclio functions the handler is in the form of module:function, defaults to "main:handler"
    :param source:   valid absolute path or URL to git, zip, or tar file, e.g.
        `git://github.com/mlrun/something.git`,
        `http://some/url/file.zip`
        note path source must exist on the image or exist locally when run is local
        (it is recommended to use 'function.spec.workdir' when source is a filepath instead)
    :param requirements:        a list of python packages, defaults to None
    :param requirements_file:   path to a python requirements file
    :param kfp:      reserved, flag indicating running within kubeflow pipeline

    :return: function object
    """
    # don't override given dict
    if runtime and isinstance(runtime, dict):
        runtime = deepcopy(runtime)
    kind, runtime = _process_runtime(command, runtime, kind)
    command = get_in(runtime, "spec.command", command)
    name = name or get_in(runtime, "metadata.name", "")

    if not kind and not command:
        runner = HandlerRuntime()
    else:
        if kind in ["", "local"] and command:
            runner = LocalRuntime.from_dict(runtime)
        elif kind in RuntimeKinds.all():
            runner = get_runtime_class(kind).from_dict(runtime)
        else:
            supported_runtimes = ",".join(RuntimeKinds.all())
            raise Exception(
                f"Unsupported runtime ({kind}) or missing command, supported runtimes: {supported_runtimes}"
            )

    if not name:
        if command and kind not in [RuntimeKinds.remote]:
            name, _ = path.splitext(path.basename(command))
        else:
            name = "mlrun-" + uuid.uuid4().hex[0:6]

    # make sure function name is valid
    name = mlrun.utils.helpers.normalize_name(name)
    mlrun.utils.helpers.validate_function_name(name)

    runner.metadata.name = name
    runner.metadata.project = (
        runner.metadata.project or project or mlconf.active_project
    )
    if tag:
        runner.metadata.tag = tag
    if image:
        if kind in ["", "handler", "local"]:
            raise ValueError(
                "image should only be set with containerized "
                "runtimes (job, mpijob, spark, ..), set kind=.."
            )
        runner.spec.image = image
    if args:
        runner.spec.args = args
    runner.kfp = kfp
    if mode:
        runner.spec.mode = mode
    if source:
        runner.spec.build.source = source
    if handler:
        if kind in RuntimeKinds.handlerless_runtimes():
            raise MLRunInvalidArgumentError(
                f"Handler is not supported for {kind} runtime"
            )
        elif kind in RuntimeKinds.nuclio_runtimes():
            runner.spec.function_handler = handler
        else:
            runner.spec.default_handler = handler

    if requirements or requirements_file:
        runner.with_requirements(
            requirements,
            requirements_file=requirements_file,
            prepare_image_for_deploy=False,
        )

    runner.prepare_image_for_deploy()

    return runner


def _process_runtime(command, runtime, kind):
    if runtime and hasattr(runtime, "to_dict"):
        runtime = runtime.to_dict()
    if runtime and isinstance(runtime, dict):
        kind = kind or runtime.get("kind", "")
        command = command or get_in(runtime, "spec.command", "")
    if "://" in command and command.startswith("http"):
        kind = kind or RuntimeKinds.remote
    if not runtime:
        runtime = {}
    update_in(runtime, "spec.command", command)
    if kind != RuntimeKinds.remote:
        if command:
            update_in(runtime, "spec.command", command)
    else:
        update_in(runtime, "spec.function_kind", "mlrun")
    return kind, runtime


def code_to_function(
    name: Optional[str] = "",
    project: Optional[str] = "",
    tag: Optional[str] = "",
    filename: Optional[str] = "",
    handler: Optional[str] = "",
    kind: Optional[str] = "",
    image: Optional[str] = None,
    code_output: Optional[str] = "",
    embed_code: bool = True,
    description: Optional[str] = "",
    requirements: Optional[list[str]] = None,
    categories: Optional[list[str]] = None,
    labels: Optional[dict[str, str]] = None,
    with_doc: Optional[bool] = True,
    ignored_tags: Optional[str] = None,
    requirements_file: Optional[str] = "",
) -> Union[
    MpiRuntimeV1,
    RemoteRuntime,
    ServingRuntime,
    DaskCluster,
    KubejobRuntime,
    LocalRuntime,
    Spark3Runtime,
    RemoteSparkRuntime,
    DatabricksRuntime,
    ApplicationRuntime,
]:
    """Convenience function to insert code and configure an mlrun runtime.

    Easiest way to construct a runtime type object. Provides the most often
    used configuration options for all runtimes as parameters.

    Instantiated runtimes are considered 'functions' in mlrun, but they are
    anything from nuclio functions to generic kubernetes pods to spark jobs.
    Functions are meant to be focused, and as such limited in scope and size.
    Typically, a function can be expressed in a single python module with
    added support from custom docker images and commands for the environment.
    The returned runtime object can be further configured if more
    customization is required.

    One of the most important parameters is 'kind'. This is what is used to
    specify the chosen runtimes. The options are:

    - local: execute a local python or shell script
    - job: insert the code into a Kubernetes pod and execute it
    - nuclio: insert the code into a real-time serverless nuclio function
    - serving: insert code into orchestrated nuclio function(s) forming a DAG
    - dask: run the specified python code / script as Dask Distributed job
    - mpijob: run distributed Horovod jobs over the MPI job operator
    - spark: run distributed Spark job using Spark Kubernetes Operator
    - remote-spark: run distributed Spark job on remote Spark service
    - databricks: run code on Databricks cluster (python scripts, Spark etc.)
    - application: run a long living application (e.g. a web server, UI, etc.)

    Learn more about :doc:`../../concepts/functions-overview`

    :param name:         function name, typically best to use hyphen-case
    :param project:      project used to namespace the function, defaults to the active project
    :param tag:          function tag to track multiple versions of the same function, defaults to 'latest'
    :param filename:     path to .py/.ipynb file, defaults to current jupyter notebook
    :param handler:      The default function handler to call for the job or nuclio function, in batch functions
                         (job, mpijob, ..) the handler can also be specified in the `.run()` command, when not specified
                         the entire file will be executed (as main).
                         for nuclio functions the handler is in the form of module:function, defaults to 'main:handler'
    :param kind:         function runtime type string - nuclio, job, etc. (see docstring for all options)
    :param image:        base docker image to use for building the function container, defaults to None
    :param code_output:  specify '.' to generate python module from the current jupyter notebook
    :param embed_code:   indicates whether or not to inject the code directly into the function runtime spec,
                         defaults to True
    :param description:  short function description, defaults to ''
    :param requirements: a list of python packages
    :param requirements_file: path to a python requirements file
    :param categories:   list of categories for mlrun Function Hub, defaults to None
    :param labels:       name/value pairs dict to tag the function with useful metadata, defaults to None
    :param with_doc:     indicates whether to document the function parameters, defaults to True
    :param ignored_tags: notebook cells to ignore when converting notebooks to py code (separated by ';')

    :return:
        pre-configured function object from a mlrun runtime class

    example::

        import mlrun

        # create job function object from notebook code and add doc/metadata
        fn = mlrun.code_to_function(
            "file_utils",
            kind="job",
            handler="open_archive",
            image="mlrun/mlrun",
            description="this function opens a zip archive into a local/mounted folder",
            categories=["fileutils"],
            labels={"author": "me"},
        )

    example::

        import mlrun
        from pathlib import Path

        # create file
        Path("mover.py").touch()

        # create nuclio function object from python module call mover.py
        fn = mlrun.code_to_function(
            "nuclio-mover",
            kind="nuclio",
            filename="mover.py",
            image="python:3.9",
            description="this function moves files from one system to another",
            requirements=["pandas"],
            labels={"author": "me"},
        )

    """
    filebase, _ = path.splitext(path.basename(filename))
    ignored_tags = ignored_tags or mlconf.ignored_notebook_tags

    def add_name(origin, name=""):
        name = filename or (name + ".ipynb")
        if not origin:
            return name
        return f"{origin}:{name}"

    def update_common(fn, spec):
        fn.spec.image = image or get_in(spec, "spec.image", "")
        fn.spec.filename = filename or get_in(spec, "spec.filename", "")
        fn.spec.build.base_image = get_in(spec, "spec.build.baseImage")
        fn.spec.build.commands = get_in(spec, "spec.build.commands")
        fn.spec.build.secret = get_in(spec, "spec.build.secret")

        if requirements or requirements_file:
            fn.with_requirements(requirements, requirements_file=requirements_file)

        if embed_code:
            fn.spec.build.functionSourceCode = get_in(
                spec, "spec.build.functionSourceCode"
            )

        if fn.kind != "local":
            fn.spec.env = get_in(spec, "spec.env")
            for vol in get_in(spec, "spec.volumes", []):
                fn.spec.volumes.append(vol.get("volume"))
                fn.spec.volume_mounts.append(vol.get("volumeMount"))

        fn.spec.description = description
        fn.metadata.project = project or mlconf.active_project
        fn.metadata.tag = tag
        fn.metadata.categories = categories
        fn.metadata.labels = labels or fn.metadata.labels

    if (
        not embed_code
        and not code_output
        and (not filename or filename.endswith(".ipynb"))
    ):
        raise ValueError(
            "A valid code file must be specified "
            "when not using the embed_code option"
        )

    if kind == RuntimeKinds.databricks and not embed_code:
        raise ValueError("Databricks tasks only support embed_code=True")

    if kind == RuntimeKinds.application:
        raise MLRunInvalidArgumentError(
            "Embedding a code file is not supported for application runtime. "
            "Code files should be specified via project/function source."
        )

    is_nuclio, sub_kind = RuntimeKinds.resolve_nuclio_sub_kind(kind)
    code_origin = add_name(add_code_metadata(filename), name)

    name, spec, code = nuclio.build_file(
        filename,
        name=name,
        handler=handler or "handler",
        kind=sub_kind,
        ignored_tags=ignored_tags,
    )

    mlrun.utils.helpers.validate_function_name(name)

    spec["spec"]["env"].append(
        {
            "name": "MLRUN_HTTPDB__NUCLIO__EXPLICIT_ACK",
            "value": mlrun.mlconf.httpdb.nuclio.explicit_ack,
        }
    )
    spec_kind = get_in(spec, "kind", "")
    if not kind and spec_kind not in ["", "Function"]:
        kind = spec_kind.lower()

        # if its a nuclio sub kind, redo nb parsing
        is_nuclio, sub_kind = RuntimeKinds.resolve_nuclio_sub_kind(kind)
        if is_nuclio:
            name, spec, code = nuclio.build_file(
                filename,
                name=name,
                handler=handler or "handler",
                kind=sub_kind,
                ignored_tags=ignored_tags,
            )

    if code_output:
        if code_output == ".":
            code_output = name + ".py"
        if filename == "" or filename.endswith(".ipynb"):
            with open(code_output, "w") as fp:
                fp.write(code)
        else:
            raise ValueError("code_output option is only used with notebooks")

    if is_nuclio:
        mlrun.utils.helpers.validate_single_def_handler(
            function_kind=sub_kind, code=code
        )

        runtime = RuntimeKinds.resolve_nuclio_runtime(kind, sub_kind)
        # default_handler is only used in :mlrun sub kind, determine the handler to invoke in function.run()
        runtime.spec.default_handler = handler if sub_kind == "mlrun" else ""
        runtime.spec.function_handler = (
            handler if handler and ":" in handler else get_in(spec, "spec.handler")
        )
        if not embed_code:
            runtime.spec.source = filename
        nuclio_runtime = get_in(spec, "spec.runtime")
        if nuclio_runtime and not nuclio_runtime.startswith("py"):
            runtime.spec.nuclio_runtime = nuclio_runtime
        if not name:
            raise ValueError("Missing required parameter: name")
        runtime.metadata.name = name
        runtime.spec.build.code_origin = code_origin
        runtime.spec.build.origin_filename = filename or (name + ".ipynb")
        update_common(runtime, spec)

        return runtime

    if kind is None or kind in ["", "Function"]:
        raise ValueError("please specify the function kind")
    elif kind in RuntimeKinds.all():
        runtime = get_runtime_class(kind)()
    else:
        raise ValueError(f"unsupported runtime ({kind})")

    name, spec, code = nuclio.build_file(filename, name=name, ignored_tags=ignored_tags)

    if not name:
        raise ValueError("name must be specified")

    h = get_in(spec, "spec.handler", "").split(":")
    runtime.handler = h[0] if len(h) <= 1 else h[1]
    runtime.metadata = get_in(spec, "spec.metadata")
    runtime.metadata.name = name
    build = runtime.spec.build
    build.code_origin = code_origin
    build.origin_filename = filename or (name + ".ipynb")
    build.extra = get_in(spec, "spec.build.extra")
    build.extra_args = get_in(spec, "spec.build.extra_args")
    build.builder_env = get_in(spec, "spec.build.builder_env")
    if not embed_code:
        if code_output:
            runtime.spec.command = code_output
        else:
            runtime.spec.command = filename

    build.image = get_in(spec, "spec.build.image")
    update_common(runtime, spec)
    runtime.prepare_image_for_deploy()

    if with_doc:
        update_function_entry_points(runtime, code)
    runtime.spec.default_handler = handler
    return runtime


def _run_pipeline(
    pipeline,
    arguments=None,
    project=None,
    experiment=None,
    run=None,
    namespace=None,
    artifact_path=None,
    ops=None,
    url=None,
    cleanup_ttl=None,
    timeout=60,
):
    """remote KubeFlow pipeline execution

    Submit a workflow task to KFP via mlrun API service

    :param pipeline:   KFP pipeline function or path to .yaml/.zip pipeline file
    :param arguments:  pipeline arguments
    :param project:    name of project
    :param experiment: experiment name
    :param run:        optional, run name
    :param namespace:  Kubernetes namespace (if not using default)
    :param url:        optional, url to mlrun API service
    :param artifact_path:  target location/url for mlrun artifacts
    :param ops:        additional operators (.apply() to all pipeline functions)
    :param cleanup_ttl:
                       pipeline cleanup ttl in secs (time to wait after workflow completion, at which point the
                       workflow and all its resources are deleted)

    :returns: kubeflow pipeline id
    """
    mldb = mlrun.db.get_run_db(url)
    if mldb.kind != "http":
        raise ValueError(
            "run pipeline require access to remote api-service"
            ", please set the dbpath url"
        )

    pipeline_run_id = mldb.submit_pipeline(
        project,
        pipeline,
        arguments,
        experiment=experiment,
        run=run,
        namespace=namespace,
        ops=ops,
        artifact_path=artifact_path,
        cleanup_ttl=cleanup_ttl,
        timeout=timeout,
    )
    logger.info(f"Pipeline run id={pipeline_run_id}, check UI for progress")
    return pipeline_run_id


def retry_pipeline(
    run_id: str,
    project: str,
) -> typing.Union[str, dict[str, str]]:
    """Retry a pipeline run.

    This function retries a previously executed pipeline run using the specified run ID. If the run is not in a
    retryable state, a new run is created as a clone of the original run.

    :param run_id: ID of the pipeline run to retry.
    :param project: name of the project associated with the pipeline run.

    :returns: ID of the retried pipeline run or the ID of a cloned run if the original run is not retryable.
    :raises ValueError: If access to the remote API service is not available.
    """
    mldb = mlrun.db.get_run_db()
    if mldb.kind != "http":
        raise ValueError(
            "Retrying a pipeline requires access to remote API service. "
            "Please set the dbpath URL."
        )

    # Invoke retry pipeline run. Depending on the context, this call returns either:
    # 1. A simple string of a workflow-id, for direct retries or non-remote workflows, or
    # 2. A dict payload representing a WorkflowResponse when rerunning remote workflows.
    rerun_response = mldb.retry_pipeline(
        run_id=run_id,
        project=project,
    )
    if isinstance(rerun_response, str):
        pipeline_run_id = rerun_response
    else:
        rerun_response = mlrun.common.schemas.WorkflowResponse(**rerun_response)

        def _fetch_workflow_id():
            rerun = mldb.read_run(rerun_response.run_id, project)
            workflow_id = rerun["metadata"]["labels"].get("workflow-id")
            if not workflow_id:
                raise mlrun.errors.MLRunRuntimeError("workflow-id label not set yet")
            return workflow_id

        pipeline_run_id = mlrun.utils.helpers.retry_until_successful(
            backoff=3,
            timeout=int(mlrun.mlconf.workflows.timeouts.remote),
            logger=logger,
            verbose=False,
            _function=_fetch_workflow_id,
        )

    if pipeline_run_id == run_id:
        logger.info(
            f"Retried pipeline run ID={pipeline_run_id}, check UI for progress."
        )
    else:
        logger.info(
            f"Copy of pipeline {run_id} was retried as run ID={pipeline_run_id}, check UI for progress."
        )
    return pipeline_run_id


def terminate_pipeline(
    run_id: str,
    project: str,
) -> str:
    """Terminate a pipeline run.

    This function terminates a running pipeline with the specified run ID. If the run is not in a
    terminable state, an error is raised.

    :param run_id: ID of the pipeline run to terminate.
    :param project: name of the project associated with the pipeline run.

    :returns: ID of the terminate pipeline run background task.
    :raises ValueError: If access to the remote API service is not available.
    """
    mldb = mlrun.db.get_run_db()
    if mldb.kind != "http":
        raise ValueError(
            "Terminating a pipeline requires access to remote API service. "
            "Please set the dbpath URL."
        )

    pipeline_run_task = mldb.terminate_pipeline(
        run_id=run_id,
        project=project,
    )
    return pipeline_run_task["metadata"]["id"]


def wait_for_pipeline_completion(
    run_id,
    timeout=60 * 60,
    expected_statuses: Optional[list[str]] = None,
    namespace=None,
    remote=True,
    project: Optional[str] = None,
):
    """Wait for Pipeline status, timeout in sec

    :param run_id:     id of pipelines run
    :param timeout:    wait timeout in sec
    :param expected_statuses:  list of expected statuses, one of [ Succeeded | Failed | Skipped | Error ], by default
                               [ Succeeded ]
    :param namespace:  k8s namespace if not default
    :param remote:     read kfp data from mlrun service (default=True)
    :param project:    the project of the pipeline

    :return: kfp run dict
    """
    if expected_statuses is None:
        expected_statuses = [RunStatuses.succeeded]
    namespace = namespace or mlconf.namespace
    logger.debug(
        f"Waiting for run completion."
        f" run_id: {run_id},"
        f" project: {project},"
        f" expected_statuses: {expected_statuses},"
        f" timeout: {timeout},"
        f" remote: {remote},"
        f" namespace: {namespace}"
    )

    if remote:
        mldb = mlrun.db.get_run_db()

        dag_display_id = create_ipython_display()

        def _wait_for_pipeline_completion():
            pipeline = mldb.get_pipeline(run_id, namespace=namespace, project=project)
            pipeline_status = pipeline["run"]["status"]
            show_kfp_run(pipeline, dag_display_id=dag_display_id, with_html=False)
            if pipeline_status not in RunStatuses.stable_statuses():
                logger.debug(
                    "Waiting for pipeline completion",
                    run_id=run_id,
                    status=pipeline_status,
                )
                raise RuntimeError("pipeline run has not completed yet")

            return pipeline

        if mldb.kind != "http":
            raise ValueError(
                "get pipeline requires access to remote api-service"
                ", set the dbpath url"
            )

        resp = retry_until_successful(
            10,
            timeout,
            logger,
            False,
            _wait_for_pipeline_completion,
        )
    else:
        client = mlrun_pipelines.utils.get_client(
            logger=logger,
            namespace=namespace,
        )
        resp = client.wait_for_run_completion(run_id, timeout)
        if resp:
            resp = resp.to_dict()
            resp = format_summary_from_kfp_run(resp)
        show_kfp_run(resp)

    status = resp["run"]["status"] if resp else "unknown"
    message = resp["run"].get("message", "") if resp else ""
    if expected_statuses:
        if status not in expected_statuses:
            raise RuntimeError(
                f"Pipeline run status {status}{', ' + message if message else ''}"
            )

    logger.debug(
        f"Finished waiting for pipeline completion."
        f" run_id: {run_id},"
        f" status: {status},"
        f" message: {message},"
        f" namespace: {namespace}"
    )

    return resp


def get_pipeline(
    run_id,
    namespace=None,
    format_: Union[
        str, mlrun.common.formatters.PipelineFormat
    ] = mlrun.common.formatters.PipelineFormat.summary,
    project: Optional[str] = None,
    remote: bool = True,
):
    """Get Pipeline status

    :param run_id:     id of pipelines run
    :param namespace:  k8s namespace if not default
    :param format_:    Format of the results. Possible values are:
            - ``summary`` (default value) - Return summary of the object data.
            - ``full`` - Return full pipeline object.
    :param project:    the project of the pipeline run
    :param remote:     read kfp data from mlrun service (default=True)

    :return: kfp run
    """
    namespace = namespace or mlconf.namespace
    if remote:
        mldb = mlrun.db.get_run_db()
        if mldb.kind != "http":
            raise ValueError(
                "get pipeline require access to remote api-service"
                ", please set the dbpath url"
            )

        resp = mldb.get_pipeline(
            run_id, namespace=namespace, format_=format_, project=project
        )

    else:
        client = mlrun_pipelines.utils.get_client(
            logger=logger,
            namespace=namespace,
        )
        resp = client.get_run(run_id)
        if resp:
            resp = resp.to_dict()
            if (
                not format_
                or format_ == mlrun.common.formatters.PipelineFormat.summary.value
            ):
                resp = mlrun.common.formatters.PipelineFormat.format_obj(resp, format_)

    show_kfp_run(resp)
    return resp


def list_pipelines(
    full=False,
    page_token="",
    page_size=None,
    sort_by="",
    filter_="",
    namespace=None,
    project="*",
    format_: mlrun.common.formatters.PipelineFormat = mlrun.common.formatters.PipelineFormat.metadata_only,
) -> tuple[int, Optional[int], list[dict]]:
    """List pipelines

    :param full:       Deprecated, use `format_` instead. if True will set `format_` to full, otherwise `format_` will
                       be used
    :param page_token: A page token to request the next page of results. The token is acquired from the nextPageToken
                       field of the response from the previous call or can be omitted when fetching the first page.
    :param page_size:  The number of pipelines to be listed per page. If there are more pipelines than this number, the
                       response message will contain a nextPageToken field you can use to fetch the next page.
    :param sort_by:    Can be format of "field_name", "field_name asc" or "field_name desc" (Example, "name asc"
                       or "id desc"). Ascending by default.
    :param filter_:    A url-encoded, JSON-serialized Filter protocol buffer, see:
                       [filter.proto](https://github.com/kubeflow/pipelines/ blob/master/backend/api/filter.proto).
    :param namespace:  Kubernetes namespace if other than default
    :param project:    Can be used to retrieve only specific project pipelines. "*" for all projects. Note that
                       filtering by project can't be used together with pagination, sorting, or custom filter.
    :param format_:    Control what will be returned (full/metadata_only/name_only)
    """
    if full:
        format_ = mlrun.common.formatters.PipelineFormat.full
    run_db = mlrun.db.get_run_db()
    pipelines = run_db.list_pipelines(
        project, namespace, sort_by, page_token, filter_, format_, page_size
    )
    return pipelines.total_size, pipelines.next_page_token, pipelines.runs


def get_object(url, secrets=None, size=None, offset=0, db=None):
    """get mlrun dataitem body (from path/url)"""
    stores = store_manager.set(secrets, db=db)
    return stores.object(url=url).get(size, offset)


def get_dataitem(url, secrets=None, db=None) -> "DataItem":
    """get mlrun dataitem object (from path/url)"""
    stores = store_manager.set(secrets, db=db)
    return stores.object(url=url)


def get_model_provider(
    url,
    secrets=None,
    db=None,
    default_invoke_kwargs: Optional[dict] = None,
    raise_missing_schema_exception=True,
) -> ModelProvider:
    """get mlrun dataitem object (from path/url)"""
    #  without caching secrets
    store_manager.set(db=db)
    return store_manager.model_provider_object(
        url=url,
        default_invoke_kwargs=default_invoke_kwargs,
        raise_missing_schema_exception=raise_missing_schema_exception,
        secrets=secrets,
    )


def download_object(url, target, secrets=None):
    """download mlrun dataitem (from path/url to target path)"""
    stores = store_manager.set(secrets)
    stores.object(url=url).download(target_path=target)


def wait_for_runs_completion(
    runs: typing.Union[list, typing.ValuesView], sleep=3, timeout=0, silent=False
):
    """wait for multiple runs to complete

    Note: need to use `watch=False` in `.run()` so the run will not wait for completion

    example::

        # run two training functions in parallel and wait for the results
        inputs = {"dataset": cleaned_data}
        run1 = train.run(
            name="train_lr",
            inputs=inputs,
            watch=False,
            params={
                "model_pkg_class": "sklearn.linear_model.LogisticRegression",
                "label_column": "label",
            },
        )
        run2 = train.run(
            name="train_lr",
            inputs=inputs,
            watch=False,
            params={
                "model_pkg_class": "sklearn.ensemble.RandomForestClassifier",
                "label_column": "label",
            },
        )
        completed = wait_for_runs_completion([run1, run2])

    :param runs:    list of run objects (the returned values of function.run())
    :param sleep:   time to sleep between checks (in seconds)
    :param timeout: maximum time to wait in seconds (0 for unlimited)
    :param silent:  set to True for silent exit on timeout
    :return: list of completed runs
    """
    completed = []
    total_time = 0
    while True:
        running = []
        for run in runs:
            state = run.state()
            if state in mlrun.common.runtimes.constants.RunStates.terminal_states():
                completed.append(run)
            else:
                running.append(run)
        if len(running) == 0:
            break
        time.sleep(sleep)
        total_time += sleep
        if timeout and total_time > timeout:
            if silent:
                break
            raise MLRunTimeoutError("some runs did not reach terminal state on time")
        runs = running

    return completed


def _ensure_path_confined_to_base_dir(
    base_directory: str,
    relative_path: str,
    error_message_on_escape: str,
) -> str:
    """
    Join `user_supplied_relative_path` to `allowed_base_directory`, normalise the result,
    and guarantee it stays inside `allowed_base_directory`.
    """
    absolute_base_directory = path.abspath(base_directory)
    absolute_candidate_path = path.abspath(
        path.join(absolute_base_directory, relative_path)
    )
    if not absolute_candidate_path.startswith(absolute_base_directory + path.sep):
        raise ValueError(error_message_on_escape)
    return absolute_candidate_path



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import typing

import orjson

import mlrun.common.schemas
import mlrun.lists
import mlrun.utils.helpers

from .base import NotificationBase, TimedHTTPClient


class WebhookNotification(NotificationBase):
    """
    API/Client notification for sending run statuses in a http request
    """

    @classmethod
    def validate_params(cls, params):
        url = params.get("url", None)
        if not url:
            raise ValueError("Parameter 'url' is required for WebhookNotification")

    async def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        url = self.params.get("url", None)
        method = self.params.get("method", "post").lower()
        headers = self.params.get("headers", {})
        override_body = self.params.get("override_body", None)
        verify_ssl = self.params.get("verify_ssl", None)

        request_body = {
            "message": message,
            "severity": severity,
        }

        if runs:
            request_body["runs"] = runs

        if alert:
            request_body["name"] = alert.name
            request_body["project"] = alert.project
            request_body["severity"] = alert.severity
            if alert.summary:
                request_body["summary"] = mlrun.utils.helpers.format_alert_summary(
                    alert, event_data
                )

            if event_data:
                request_body["value"] = event_data.value_dict
                request_body["id"] = event_data.entity.ids[0]

        if custom_html:
            request_body["custom_html"] = custom_html

        if override_body:
            request_body = self._serialize_runs_in_request_body(override_body, runs)

        # Specify the `verify_ssl` parameter value only for HTTPS urls.
        # The `ClientSession` allows using `ssl=None` for the default SSL check,
        # and `ssl=False` to skip SSL certificate validation.
        # We maintain the default as `None`, so if the user sets `verify_ssl=True`,
        # we automatically handle it as `ssl=None` for their convenience.
        verify_ssl = verify_ssl and None if url.startswith("https") else None

        async with TimedHTTPClient().session(json_serialize=self._encoder) as session:
            response = await getattr(session, method)(
                url,
                headers=headers,
                json=request_body,
                ssl=verify_ssl,
            )
            response.raise_for_status()

    @staticmethod
    def _serialize_runs_in_request_body(override_body, runs):
        runs = runs or []

        def parse_runs():
            parsed_runs = []
            for run in runs:
                if hasattr(run, "to_dict"):
                    run = run.to_dict()
                if isinstance(run, dict):
                    parsed_run = {
                        "project": run["metadata"]["project"],
                        "name": run["metadata"]["name"],
                        "status": {"state": run["status"]["state"]},
                    }
                    if host := run["metadata"].get("labels", {}).get("host", ""):
                        parsed_run["host"] = host
                    if error := run["status"].get("error"):
                        parsed_run["status"]["error"] = error
                    elif results := run["status"].get("results"):
                        parsed_run["status"]["results"] = results
                    parsed_runs.append(parsed_run)
            return str(parsed_runs)

        if isinstance(override_body, dict):
            for key, value in override_body.items():
                if not isinstance(value, str):
                    # If the value is not a string, we don't want to parse it
                    continue
                if re.search(r"{{\s*runs\s*}}", value):
                    str_parsed_runs = parse_runs()
                    override_body[key] = re.sub(
                        r"{{\s*runs\s*}}", str_parsed_runs, value
                    )

        return override_body

    @property
    def _encoder(self):
        return lambda body: orjson.dumps(
            body,
            option=orjson.OPT_NAIVE_UTC
            | orjson.OPT_SERIALIZE_NUMPY
            | orjson.OPT_NON_STR_KEYS
            | orjson.OPT_SORT_KEYS,
        ).decode()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pandas as pd

import mlrun
import mlrun.common.constants as mlrun_constants
from mlrun.datastore.sources import get_source_from_dict, get_source_step
from mlrun.datastore.targets import (
    add_target_steps,
    get_target_driver,
    validate_target_list,
    validate_target_placement,
)

from ..data_types import InferOptions
from ..datastore.store_resources import ResourceCache
from ..runtimes import RuntimeKinds
from ..runtimes.function_reference import FunctionReference
from ..serving.server import MockEvent, create_graph_server
from ..utils import logger, normalize_name
from .feature_set import FeatureSet


def init_featureset_graph(
    source,
    featureset,
    namespace,
    targets=None,
    return_df=True,
    verbose=False,
    rows_limit=None,
):
    """create storey ingestion graph/DAG from feature set object"""

    cache = ResourceCache()
    graph = featureset.spec.graph.copy()

    # init targets (and table)
    targets = targets or []
    server = create_graph_server(graph=graph, parameters={}, verbose=verbose)
    server.init_states(context=None, namespace=namespace, resource_cache=cache)

    if graph.engine != "sync":
        # todo: support rows_limit it storey sources
        _add_data_steps(
            graph,
            cache,
            featureset,
            targets=targets,
            source=source,
            return_df=return_df,
            context=server.context,
        )
        server.init_object(namespace)
        return graph.wait_for_completion()
    else:
        # for initialize all the validators of the feature set
        cache.cache_resource(featureset.uri, featureset, True)

    server.init_object(namespace)

    # if the source is a dataframe iterator we load/write it in chunks
    chunk_id = 0
    if hasattr(source, "to_dataframe"):
        if source.is_iterator():
            chunk_id = 1
            chunks = source.to_dataframe()
        else:
            chunks = [source.to_dataframe()]
    elif not hasattr(source, "to_csv"):
        raise mlrun.errors.MLRunInvalidArgumentError("illegal source")
    else:
        chunks = [source]

    entity_columns = list(featureset.spec.entities.keys())
    key_fields = entity_columns if entity_columns else None

    sizes = [0] * len(targets)
    result_dfs = []
    total_rows = 0
    targets = [get_target_driver(target, featureset) for target in targets]
    if featureset.spec.passthrough:
        targets = [target for target in targets if not target.is_offline]
    for chunk in chunks:
        event = MockEvent(body=chunk)
        if len(featureset.spec.entities) and isinstance(event.body, pd.DataFrame):
            # set the entities to be the indexes of the df
            event.body = entities_to_index(featureset, event.body)

        df = server.run(event, get_body=True)
        if df is not None:
            for i, target in enumerate(targets):
                size = target.write_dataframe(
                    df,
                    key_column=key_fields,
                    timestamp_key=featureset.spec.timestamp_key,
                    chunk_id=chunk_id,
                )
                if size:
                    sizes[i] += size
        chunk_id += 1
        result_dfs.append(df)
        total_rows += df.shape[0]
        if rows_limit and total_rows >= rows_limit:
            break

    for i, target in enumerate(targets):
        target_status = target.update_resource_status("ready", size=sizes[i])
        if verbose:
            logger.info(f"wrote target: {target_status}")

    result_df = pd.concat(result_dfs)
    return result_df.head(rows_limit)


def featureset_initializer(server):
    """graph server hook to initialize feature set ingestion graph/DAG"""

    context = server.context
    cache = server.resource_cache
    featureset, source, targets, _, _ = context_to_ingestion_params(context)

    graph = featureset.spec.graph.copy()
    _add_data_steps(
        graph,
        cache,
        featureset,
        targets=targets,
        source=source,
        context=context,
    )
    featureset.save()
    server.graph = graph


def run_spark_graph(df, featureset, namespace, spark):
    """run spark (sync) pipeline"""
    cache = ResourceCache()
    graph = featureset.spec.graph.copy()
    if graph.engine != "sync":
        raise mlrun.errors.MLRunInvalidArgumentError("spark must use sync graph")
    for step_dict in graph.steps.values():
        if step_dict.class_name in [
            "mlrun.feature_store.steps.FeaturesetValidator",
            "mlrun.feature_store.steps.SetEventMetadata",
        ]:
            raise mlrun.errors.MLRunRuntimeError(
                f"{step_dict.class_name} is not supported for spark engine."
            )
    server = create_graph_server(graph=graph, parameters={})
    server.init_states(context=None, namespace=namespace, resource_cache=cache)
    server.init_object(namespace)
    server.context.spark = spark
    event = MockEvent(body=df)
    return server.run(event, get_body=True)


def context_to_ingestion_params(context):
    """extract the ingestion task params from job/serving context"""

    featureset_uri = context.get_param("featureset")
    featureset = context.get_store_resource(featureset_uri)
    infer_options = context.get_param("infer_options", InferOptions.Null)

    source = context.get_param("source")
    if source:
        source = get_source_from_dict(source)
    elif featureset.spec.source.to_dict():
        source = get_source_from_dict(featureset.spec.source.to_dict())
    overwrite = context.get_param("overwrite", None)

    targets = context.get_param("targets", None)
    if not targets:
        targets = featureset.spec.targets
    targets = [get_target_driver(target, featureset) for target in targets]
    return featureset, source, targets, infer_options, overwrite


def _add_data_steps(
    graph, cache, featureset, targets, source, return_df=False, context=None
):
    _, default_final_step, _ = graph.check_and_process_graph(allow_empty=True)
    validate_target_list(targets=targets)
    validate_target_placement(graph, default_final_step, targets)
    cache.cache_resource(featureset.uri, featureset, True)
    table = add_target_steps(
        graph, featureset, targets, to_df=return_df, final_step=default_final_step
    )
    if table:
        cache.cache_table(featureset.uri, table, True)

    entity_columns = list(featureset.spec.entities.keys())
    key_fields = entity_columns if entity_columns else None

    if source is not None:
        source = get_source_step(
            source,
            key_fields=key_fields,
            time_field=featureset.spec.timestamp_key,
            context=context,
        )
    graph.set_flow_source(source)


def run_ingestion_job(name, featureset, run_config, schedule=None, spark_service=None):
    name = normalize_name(name or f"{featureset.metadata.name}-ingest-job")
    use_spark = featureset.spec.engine == "spark"
    spark_runtimes = [RuntimeKinds.remotespark, RuntimeKinds.spark]

    default_kind = RuntimeKinds.remotespark if use_spark else RuntimeKinds.job

    if not run_config.function:
        function_ref = featureset.spec.function.copy()
        if function_ref.is_empty():
            function_ref = FunctionReference(name=name, kind=default_kind)
        if not function_ref.url:
            function_ref.code = (function_ref.code or "") + _default_job_handler
        run_config.function = function_ref
        run_config.handler = "handler"
    elif run_config.function.kind == RuntimeKinds.spark and spark_service is not None:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Spark operator jobs do not support standalone spark submission"
        )

    image = None if use_spark else mlrun.mlconf.feature_store.default_job_image
    function = run_config.to_function(default_kind, image)
    if use_spark and function.kind not in spark_runtimes:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "ingest with spark engine require spark function kind"
        )

    function.metadata.project = featureset.metadata.project
    function.metadata.name = function.metadata.name or name

    if not use_spark and not function.spec.image:
        raise mlrun.errors.MLRunInvalidArgumentError("function image must be specified")

    if use_spark and function.kind == RuntimeKinds.remotespark and not run_config.local:
        if not spark_service:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Remote spark ingestion requires the spark service name to be provided"
            )
        else:
            function.with_spark_service(spark_service=spark_service)

    task = mlrun.new_task(
        name=name,
        params=run_config.parameters,
        handler=run_config.handler,
        out_path=featureset.spec.output_path,
    )
    task.spec.secret_sources = run_config.secret_sources
    task.set_label(
        mlrun_constants.MLRunInternalLabels.job_type, "feature-ingest"
    ).set_label("feature-set", featureset.uri)
    if run_config.owner:
        task.set_label(
            mlrun_constants.MLRunInternalLabels.owner, run_config.owner
        ).set_label(mlrun_constants.MLRunInternalLabels.v3io_user, run_config.owner)

    # set run UID and save in the feature set status (linking the features et to the job)
    task.metadata.uid = uuid.uuid4().hex
    featureset.status.run_uri = task.metadata.uid
    featureset.save()

    # when running in server side we want to set the function db connection to the actual DB and not to use the httpdb
    function.set_db_connection(featureset._get_run_db())

    run = function.run(
        task,
        schedule=schedule,
        local=run_config.local,
        watch=run_config.watch,
        auth_info=run_config.auth_info,
    )
    if run_config.watch:
        featureset.reload()
    return run


def entities_to_index(featureset: FeatureSet, data: pd.DataFrame) -> pd.DataFrame:
    entities_names = [
        ent.name for ent in featureset.spec.entities if ent.name in data.columns
    ]

    if len(entities_names) > 0:
        drop_columns = []
        add_indexes = []
        for ent_name in entities_names:
            if ent_name in data.index.names:
                drop_columns.append(ent_name)
            else:
                add_indexes.append(ent_name)

        # drop duplicate columns and indexes
        data = data.drop(drop_columns)

        # append or reset index (append if index is not default)
        append = data.index.names[0] is not None
        data = data.set_index(add_indexes, append=append)

    return data


_default_job_handler = """
from mlrun.feature_store.api import ingest
def handler(context):
    ingest(mlrun_context=context)
"""



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


# TODO: From python 3.11 StrEnum is built-in and this will not be needed
class StrEnum(str, enum.Enum):
    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


# Partial backport from Python 3.11
# https://docs.python.org/3/library/http.html#http.HTTPMethod
class HTTPMethod(StrEnum):
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PATCH = "PATCH"
    PUT = "PUT"


class Operation(StrEnum):
    ADD = "add"
    REMOVE = "remove"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Don't remove this, used by sphinx documentation
__all__ = ["get_model", "update_model", "DocumentLoaderSpec", "MLRunLoader"]

from .base import (
    Artifact,
    ArtifactMetadata,
    ArtifactSpec,
    DirArtifact,
    get_artifact_meta,
)
from .dataset import DatasetArtifact, TableArtifact, update_dataset_meta
from .document import DocumentArtifact, DocumentLoaderSpec, MLRunLoader
from .llm_prompt import LLMPromptArtifact, LLMPromptArtifactSpec
from .manager import (
    ArtifactManager,
    ArtifactProducer,
    artifact_types,
    dict_to_artifact,
)
from .model import ModelArtifact, get_model, update_model
from .plots import PlotArtifact, PlotlyArtifact



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Optional, Union

import sklearn
from sklearn.preprocessing import LabelBinarizer

import mlrun.errors

from .._ml_common import MLUtils
from .metric import Metric
from .utils import SKLearnTypes


class MetricsLibrary(ABC):
    """
    Static class for getting and parsing metrics to use in training and evaluation of SciKit-Learn.
    """

    # A constant name for the context parameter to use for passing a metrics configuration:
    CONTEXT_PARAMETER = "_metrics"

    # A keyword to add in case the metric is based on predictions probabilities (not final predictions):
    _NEED_PROBABILITIES_KEYWORD = "need_probabilities"

    @classmethod
    def get_metrics(
        cls,
        metrics: Optional[
            Union[
                list[Metric],
                list[SKLearnTypes.MetricEntryType],
                dict[str, SKLearnTypes.MetricEntryType],
            ]
        ] = None,
        context: mlrun.MLClientCtx = None,
        include_default: bool = True,
        **default_kwargs,
    ) -> list[Metric]:
        """
        Get metrics for a run. The metrics will be taken from the provided metrics / configuration via code, from
        provided configuration via MLRun context and if the 'include_default' is True, from the metric library's
        defaults as well.

        :param metrics:         The metrics parameter passed to the function. Can be passed as a dictionary or a list of
                                metrics.
        :param context:         A context to look in if the configuration was passed as a parameter.
        :param include_default: Whether to include the default in addition to the provided metrics. Default: True.
        :param default_kwargs:  Additional key word arguments to pass to the 'default' method of the given metrics
                                library class.

        :return: The metrics list.

        :raise MLRunInvalidArgumentError: If the metrics were not passed in a list or a dictionary.
        """
        # Set up the plans list:
        parsed_metrics = []  # type: List[Metric]

        # Get the metrics passed via context:
        if context is not None and cls.CONTEXT_PARAMETER in context.parameters:
            parsed_metrics += cls._parse(
                metrics=context.parameters.get(cls.CONTEXT_PARAMETER, None)
            )

        # Get the user's set metrics:
        if metrics is not None:
            parsed_metrics += cls._parse(metrics=metrics)

        # Get the library's default:
        if include_default:
            parsed_metrics += cls._default(**default_kwargs)

        return parsed_metrics

    @classmethod
    def _parse(
        cls,
        metrics: Union[
            list[Metric],
            list[SKLearnTypes.MetricEntryType],
            dict[str, SKLearnTypes.MetricEntryType],
        ],
    ) -> list[Metric]:
        """
        Parse the given metrics by the possible rules of the framework implementing.

        :param metrics: A collection of metrics to parse.

        :return: The parsed metrics to use in training / evaluation.
        """
        # Parse from dictionary:
        if isinstance(metrics, dict):
            return cls._from_dict(metrics_dictionary=metrics)

        # Parse from list:
        if isinstance(metrics, list):
            return cls._from_list(metrics_list=metrics)

        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The metrics are expected to be in a list or a dictionary. Received: {type(metrics)}. A metric can be a "
            f"function, callable object, name of an imported function or a module path to import the function. "
            f"Arguments can be passed as a tuple: in the following form: (metric, arguments). If used in a dictionary, "
            f"each key will be the name to use for logging the metric."
        )

    @classmethod
    def _from_list(
        cls, metrics_list: list[Union[Metric, SKLearnTypes.MetricEntryType]]
    ) -> list[Metric]:
        """
        Collect the given metrics configurations from a list. The metrics names will be chosen by the following rules:

        * If a function was given, the function's name.
        * If a callable object was given, the object's class name.
        * If a module path was given, the text after the last dot.
        * Duplicates will be counted and concatenated with their number of duplicates.

        :param metrics_list: The list of metrics. A metric can be a function, callable object, name of an imported
                             function or a module path to import the function. Arguments can be passed as a tuple:
                             (metric, arguments). To specify the metric expects probabilities (calling
                             model.predict_proba instead of model.predict), pass within the arguments:
                             { "need_probabilities": True }.

        :return: A list of metrics objects.
        """
        return [
            metric
            if isinstance(metric, Metric)
            else cls._to_metric_class(metric_entry=metric)
            for metric in metrics_list
        ]

    @classmethod
    def _from_dict(
        cls, metrics_dictionary: dict[str, SKLearnTypes.MetricEntryType]
    ) -> list[Metric]:
        """
        Collect the given metrics configurations from a dictionary.

        :param metrics_dictionary: A dictionary of metrics where each key is the metric name to be logged with and its
                                   value is the metric itself. A metric can be a function, callable object, name of an
                                   imported function or a module path to import the function. Arguments can be passed
                                   along as a tuple: (metric, arguments). To specify the metric expects probabilities
                                   (calling model.predict_proba instead of model.predict), pass within the arguments:
                                   { "need_probabilities": True }.

        :return: A list of metrics objects.
        """
        return [
            cls._to_metric_class(metric_entry=metric, metric_name=metric_name)
            for metric_name, metric in metrics_dictionary.items()
        ]

    @classmethod
    def _default(
        cls, model: SKLearnTypes.ModelType, y: SKLearnTypes.DatasetType = None
    ) -> list[Metric]:
        """
        Get the default metrics list according to the algorithm functionality.

        :param model: The model to check if its a regression model or a classification model.
        :param y:     The ground truth values to check if its multiclass and / or multi output.

        :return: The default metrics list.
        """
        # Discover the algorithm functionality of the provided model:
        algorithm_functionality = MLUtils.get_algorithm_functionality(model=model, y=y)

        # Initialize the metrics list:
        metrics = []  # type: List[Metric]

        # Add classification metrics:
        if algorithm_functionality.is_classification():
            metrics += [Metric(name="accuracy", metric=sklearn.metrics.accuracy_score)]
            if (
                algorithm_functionality.is_binary_classification()
                and algorithm_functionality.is_single_output()
            ):
                metrics += [
                    Metric(metric=sklearn.metrics.f1_score),
                    Metric(metric=sklearn.metrics.precision_score),
                    Metric(metric=sklearn.metrics.recall_score),
                ]
            if algorithm_functionality.is_multiclass_classification():
                metrics += [
                    Metric(
                        metric=sklearn.metrics.f1_score,
                        additional_arguments={"average": "macro"},
                    ),
                    Metric(
                        metric=sklearn.metrics.precision_score,
                        additional_arguments={"average": "macro"},
                    ),
                    Metric(
                        metric=sklearn.metrics.recall_score,
                        additional_arguments={"average": "macro"},
                    ),
                ]
                if algorithm_functionality.is_single_output():
                    metrics += [
                        Metric(
                            name="auc-micro",
                            metric=lambda y_true, y_pred: sklearn.metrics.roc_auc_score(
                                LabelBinarizer().fit_transform(y_true),
                                y_pred,
                                multi_class="ovo",
                                average="micro",
                            ),
                            need_probabilities=True,
                        ),
                        Metric(
                            name="auc-macro",
                            metric=sklearn.metrics.roc_auc_score,
                            additional_arguments={
                                "multi_class": "ovo",
                                "average": "macro",
                            },
                            need_probabilities=True,
                        ),
                        Metric(
                            name="auc-weighted",
                            metric=sklearn.metrics.roc_auc_score,
                            additional_arguments={
                                "multi_class": "ovo",
                                "average": "weighted",
                            },
                            need_probabilities=True,
                        ),
                    ]

        # Add regression metrics:
        if algorithm_functionality.is_regression():
            metrics += [
                Metric(metric=sklearn.metrics.mean_absolute_error),
                Metric(metric=sklearn.metrics.r2_score),
                Metric(
                    name="root_mean_squared_error",
                    metric=sklearn.metrics.mean_squared_error,
                    additional_arguments={"squared": False},
                ),
                Metric(metric=sklearn.metrics.mean_squared_error),
            ]

        # Filter out the metrics by probabilities requirement:
        if not hasattr(model, "predict_proba"):
            metrics = [metric for metric in metrics if not metric.need_probabilities]

        return metrics

    @classmethod
    def _to_metric_class(
        cls,
        metric_entry: SKLearnTypes.MetricEntryType,
        metric_name: Optional[str] = None,
    ) -> Metric:
        """
        Create a Metric instance from a user given metric entry.

        :param metric_entry: Metric entry as passed inside a list or a dictionary.
        :param metric_name:  The metric name to use (if passed from a dictionary).

        :return: The metric class instance of this entry.
        """
        # If it's a tuple, unpack it to get the additional arguments:
        if isinstance(metric_entry, tuple):
            metric, arguments = metric_entry
        else:
            metric = metric_entry
            arguments = {}

        # Check if the 'need_probabilities' attribute is given:
        if cls._NEED_PROBABILITIES_KEYWORD in arguments:
            need_probabilities = arguments[cls._NEED_PROBABILITIES_KEYWORD]
            arguments.pop(cls._NEED_PROBABILITIES_KEYWORD)
        else:
            need_probabilities = False

        # Initialize the Metric with the collected information and return:
        return Metric(
            metric=metric,
            name=metric_name,
            additional_arguments=arguments,
            need_probabilities=need_probabilities,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from os import path
from typing import Any, Optional, Union

import pandas as pd
import yaml

import mlrun
import mlrun.datastore

from ..data_types import InferOptions, get_infer_interface
from ..features import Feature
from ..model import ObjectList
from ..utils import StorePrefix, is_relative_path
from .base import Artifact, ArtifactSpec, upload_extra_data, verify_target_path

model_spec_filename = "model_spec.yaml"
MODEL_OPTIONAL_SUFFIXES = [".tar.gz", ".pkl", ".bin", ".pickle"]


class ModelArtifactSpec(ArtifactSpec):
    _dict_fields = ArtifactSpec._dict_fields + [
        "model_file",
        "metrics",
        "parameters",
        "inputs",
        "outputs",
        "framework",
        "algorithm",
        "feature_vector",
        "feature_weights",
        "feature_stats",
        "model_target_file",
        "model_url",
    ]
    _exclude_fields_from_uid_hash = ArtifactSpec._exclude_fields_from_uid_hash + [
        "metrics",
        "parameters",
        "inputs",
        "outputs",
        "feature_vector",
        "feature_weights",
        "feature_stats",
    ]

    def __init__(
        self,
        src_path=None,
        target_path=None,
        viewer=None,
        is_inline=False,
        format=None,
        size=None,
        db_key=None,
        extra_data=None,
        body=None,
        model_file=None,
        metrics=None,
        paraemeters=None,
        inputs: Optional[list[Feature]] = None,
        outputs: Optional[list[Feature]] = None,
        framework=None,
        algorithm=None,
        feature_vector=None,
        feature_weights=None,
        feature_stats=None,
        model_target_file=None,
        model_url=None,
    ):
        super().__init__(
            src_path,
            target_path,
            viewer,
            is_inline,
            format,
            size,
            db_key,
            extra_data,
            body,
        )
        self.model_file = model_file
        self.metrics = metrics or {}
        self.parameters = paraemeters or {}
        self.inputs = inputs or []
        self.outputs = outputs or []
        self.framework = framework
        self.algorithm = algorithm
        self.feature_vector = feature_vector
        self.feature_weights = feature_weights
        self.feature_stats = feature_stats
        self.model_target_file = model_target_file
        self.model_url = model_url

    @property
    def inputs(self) -> ObjectList:
        """input feature list"""
        return self._inputs

    @inputs.setter
    def inputs(self, inputs: list[Feature]) -> None:
        self._inputs = ObjectList.from_list(Feature, inputs)

    @property
    def outputs(self) -> ObjectList:
        """output feature list"""
        return self._outputs

    @outputs.setter
    def outputs(self, outputs: list[Feature]) -> None:
        self._outputs = ObjectList.from_list(Feature, outputs)

    @property
    def default_config(self):
        return self.parameters.get("default_config", {})

    @default_config.setter
    def default_config(self, default_config):
        # skip storing 'default_config' if value is empty or unset
        if default_config:
            self.parameters["default_config"] = default_config
        else:
            self.parameters.pop("default_config", None)


class ModelArtifact(Artifact):
    """ML Model artifact

    Store link to ML model file(s) along with the model metrics, parameters, schema, and stats
    """

    kind = "model"
    _store_prefix = StorePrefix.Model

    def __init__(
        self,
        key=None,
        body=None,
        format=None,
        model_file=None,
        metrics=None,
        target_path=None,
        parameters=None,
        inputs=None,
        outputs=None,
        framework=None,
        algorithm=None,
        feature_vector=None,
        feature_weights=None,
        extra_data=None,
        model_dir=None,
        model_url: Optional[str] = None,
        default_config: Optional[dict] = None,
        **kwargs,
    ):
        """
        :param key:             Artifact key or artifact class ()
        :param body:            Will use the body as the artifact content
        :param format:          Optional, format to use (e.g. csv, parquet, ..)
        :param model_file:      Path to the local model file we upload (see also model_dir)
                                or to a model file data url (e.g. `http://host/path/model.pkl`)
        :param metrics:         The key/value dict of model metrics
        :param target_path:     Absolute target path (instead of using artifact_path + local_path)
        :param parameters:      Key/value dict of model parameters
        :param inputs:          Ordered list of model input features (name, type, ..)
        :param outputs:         Ordered list of model output/result elements (name, type, ..)
        :param framework:       Name of the ML framework
        :param algorithm:       Training algorithm name
        :param feature_vector:  Feature store feature vector uri (store://feature-vectors/<project>/<name>[:tag])
        :param feature_weights: List of feature weights, one per input column
        :param extra_data:      Extra artifacts and files to log with the model.
        :param model_dir:       Path to the local dir holding the model file and extra files
        :param model_url:       Remote model url.
        :param default_config:  Default configuration for client building
                                Saved as a sub-dictionary under the parameter.
        :param kwargs:          Arguments to pass to the artifact class.
        """
        super().__init__(key, body, format=format, target_path=target_path, **kwargs)
        model_file = str(model_file or "")
        if (model_file or model_dir or body) and model_url:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Arguments 'model_file' and 'model_url' cannot be"
                " used together with 'model_file', 'model_dir' or 'body'."
            )
        if model_file and "/" in model_file:
            if model_dir:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "'model_file' cannot contain '/' (i.e., be a full path) when 'model_dir' is also specified"
                )
            model_dir = path.dirname(model_file)
            model_file = path.basename(model_file)
        self.spec.model_file = model_file
        self.spec.src_path = model_dir
        self.spec.parameters = parameters or {}
        self.spec.metrics = metrics or {}
        self.spec.inputs = inputs or []
        self.spec.outputs = outputs or []
        self.spec.extra_data = extra_data or {}
        self.spec.framework = framework
        self.spec.algorithm = algorithm
        self.spec.feature_vector = feature_vector
        self.spec.feature_weights = feature_weights
        self.spec.feature_stats = None
        self.spec.model_url = model_url
        self.default_config = default_config

    @property
    def spec(self) -> ModelArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", ModelArtifactSpec)

    @property
    def inputs(self) -> ObjectList:
        """input feature list"""
        return self.spec.inputs

    @inputs.setter
    def inputs(self, inputs: list[Feature]) -> None:
        """input feature list"""
        self.spec.inputs = inputs

    @property
    def outputs(self) -> ObjectList:
        """input feature list"""
        return self.spec.outputs

    @outputs.setter
    def outputs(self, outputs: list[Feature]) -> None:
        """input feature list"""
        self.spec.outputs = outputs

    @property
    def model_file(self):
        return self.spec.model_file

    @model_file.setter
    def model_file(self, model_file):
        self.spec.model_file = model_file

    @property
    def model_url(self):
        return self.spec.model_url

    @model_url.setter
    def model_url(self, model_url):
        self.spec.model_url = model_url

    @property
    def parameters(self):
        return self.spec.parameters

    @parameters.setter
    def parameters(self, parameters):
        self.spec.parameters = parameters

    @property
    def default_config(self):
        return self.spec.default_config

    @default_config.setter
    def default_config(self, default_config):
        self.spec.default_config = default_config

    @property
    def metrics(self):
        return self.spec.metrics

    @metrics.setter
    def metrics(self, metrics):
        self.spec.metrics = metrics

    @property
    def feature_stats(self):
        return self.spec.feature_stats

    @feature_stats.setter
    def feature_stats(self, feature_stats):
        self.spec.feature_stats = feature_stats

    @property
    def feature_vector(self):
        return self.spec.feature_vector

    @feature_vector.setter
    def feature_vector(self, feature_vector):
        self.spec.feature_vector = feature_vector

    @property
    def feature_weights(self):
        return self.spec.feature_weights

    @feature_weights.setter
    def feature_weights(self, feature_weights):
        self.spec.feature_weights = feature_weights

    @property
    def model_target_file(self):
        return self.spec.model_target_file

    @model_target_file.setter
    def model_target_file(self, model_target_file):
        self.spec.model_target_file = model_target_file

    def infer_from_df(self, df, label_columns=None, with_stats=True, num_bins=None):
        """infer inputs, outputs, and stats from provided df (training set)

        :param df:      dataframe to infer from
        :param label_columns: name of the label (target) column
        :param with_stats:    infer statistics (min, max, .. histogram)
        :param num_bins:      number of bins for histogram
        """
        subset = df
        inferer = get_infer_interface(subset)
        numeric_columns = self._extract_numeric_features(df)
        if label_columns:
            if not isinstance(label_columns, list):
                label_columns = [label_columns]
            subset = df.drop(columns=label_columns)
        inferer.infer_schema(
            subset, self.spec.inputs, {}, options=InferOptions.Features
        )
        if label_columns:
            inferer.infer_schema(
                df[label_columns],
                self.spec.outputs,
                {},
                options=InferOptions.Features,
                push_at_start=True,
            )
        if with_stats:
            self.spec.feature_stats = inferer.get_stats(
                df[numeric_columns], options=InferOptions.Histogram, num_bins=num_bins
            )

    @staticmethod
    def _extract_numeric_features(df: pd.DataFrame) -> list[Any]:
        return [col for col in df.columns if pd.api.types.is_numeric_dtype(df[col])]

    @property
    def is_dir(self):
        return True

    def before_log(self):
        if not self.spec.model_file and not self.spec.model_url:
            raise ValueError(
                "ModelArtifact must have either 'model_file' or 'model_url' attributes"
            )

        super().before_log()

        if self.spec.framework:
            self.metadata.labels = self.metadata.labels or {}
            self.metadata.labels["framework"] = self.spec.framework

    def upload(self, artifact_path: Optional[str] = None):
        """
        internal, upload to target store
        :param artifact_path: required only for when generating target_path from artifact hash
        """
        # if mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash outputs True and the user
        # didn't pass target_path explicitly, then target_path will be calculated right before uploading the artifact
        # using `resolve_<body/file>_target_hash_path`
        target_model_path = None
        if self.spec.target_path:
            target_model_path = path.join(
                self.spec.target_path, path.basename(self.spec.model_file)
            )

        target_model_path = self._upload_body_or_file(
            artifact_path, target_model_path=target_model_path
        )
        upload_extra_data(
            artifact=self, extra_data=self.spec.extra_data, artifact_path=artifact_path
        )

        spec_body = _sanitize_and_serialize_model_spec_yaml(self)
        spec_target_path = None

        if mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash:
            # resolving target_path for the model spec
            _, spec_target_path = self.resolve_body_target_hash_path(
                body=spec_body, artifact_path=artifact_path
            )

            # if mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash outputs True, then target_path
            # will point to the artifact path which is where the model and all its extra data are stored
            self.spec.target_path = (
                artifact_path + "/"
                if not artifact_path.endswith("/")
                else artifact_path
            )
            # unlike in extra_data, which stores for each key the path to the file, in target_path we store the
            # target path dir, and because we generated the target path of the model from the artifact hash,
            # the model_file doesn't represent the actual target file name of the model, so we need to update it
            self.spec.model_target_file = path.basename(target_model_path)

        spec_target_path = spec_target_path or path.join(
            self.spec.target_path, model_spec_filename
        )
        mlrun.datastore.store_manager.object(url=spec_target_path).put(spec_body)

    def _upload_body_or_file(
        self,
        artifact_path: str,
        target_model_path: Optional[str] = None,
    ):
        body = self.spec.get_body()
        if body:
            if not target_model_path:
                (
                    self.metadata.hash,
                    target_model_path,
                ) = self.resolve_body_target_hash_path(
                    body=body, artifact_path=artifact_path
                )
            self._upload_body(
                body, target=target_model_path, artifact_path=artifact_path
            )

        else:
            src_model_path = _get_src_path(self, self.spec.model_file)
            if not path.isfile(src_model_path):
                raise ValueError(f"Model file {src_model_path} not found")

            if not target_model_path:
                (
                    self.metadata.hash,
                    target_model_path,
                ) = self.resolve_file_target_hash_path(
                    source_path=src_model_path, artifact_path=artifact_path
                )

            self._upload_file(
                src_model_path,
                target_path=target_model_path,
                artifact_path=artifact_path,
            )

        return target_model_path

    def _get_file_body(self):
        body = self.spec.get_body()
        if body:
            return body
        src_model_path = _get_src_path(self, self.spec.model_file)
        if src_model_path and path.isfile(src_model_path):
            with open(src_model_path, "rb") as fp:
                return fp.read()
        target_model_path = path.join(self.spec.target_path, self.spec.model_file)
        return mlrun.get_dataitem(target_model_path).get()


def get_model(
    model_dir: Optional[
        Union[str, ModelArtifact, "mlrun.datastore.base.DataItem"]
    ] = None,
    suffix="",
) -> (str, ModelArtifact, dict):
    """Return model file, model spec object, and dictionary of extra data items

    this function will get the model file, metadata, and extra data
    the returned model file is always local, when using remote urls
    (such as v3io://, s3://, store://, ..) it will be copied locally.

    returned extra data dict (of key, DataItem objects) allow reading additional model files/objects
    e.g. use DataItem.get() or .download(target) .as_df() to read

    example::

        model_file, model_artifact, extra_data = get_model(models_path, suffix=".pkl")
        model = load(open(model_file, "rb"))
        categories = extra_data["categories"].as_df()

    :param model_dir:       model dir or artifact path (store://..) or DataItem
    :param suffix:          model filename suffix (when using a dir)

    :returns: model filename, model artifact object, extra data dict

    """
    model_file = ""
    model_spec = None
    extra_dataitems = {}
    default_suffix = ".pkl"

    if hasattr(model_dir, "artifact_url"):
        model_dir = model_dir.artifact_url
    alternative_suffix = next(
        (
            optional_suffix
            for optional_suffix in MODEL_OPTIONAL_SUFFIXES
            if isinstance(model_dir, str)
            and model_dir.lower().endswith(optional_suffix)
        ),
        None,
    )
    is_store_uri = isinstance(model_dir, str) and mlrun.datastore.is_store_uri(
        model_dir
    )
    if is_store_uri or isinstance(model_dir, ModelArtifact):
        if is_store_uri:
            model_spec, target = mlrun.datastore.store_manager.get_store_artifact(
                model_dir
            )
            verify_target_path(model_spec)
        else:
            model_spec, target = model_dir, model_dir.get_target_path()
        if not model_spec or model_spec.kind != "model":
            raise ValueError(f"store artifact ({model_dir}) is not model kind")
        # in case model_target_file is specified, use it, because that means that the actual model target path
        # in the store is different from the local model_file it was generated from
        model_file = _get_file_path(
            target, model_spec.model_target_file or model_spec.model_file
        )
        extra_dataitems = _get_extra(target, model_spec.extra_data)
        suffix = suffix or default_suffix
    elif model_dir.lower().endswith(".yaml"):
        model_spec = _load_model_spec(model_dir)
        model_file = _get_file_path(model_dir, model_spec.model_file)
        extra_dataitems = _get_extra(model_dir, model_spec.extra_data)
        suffix = suffix or default_suffix
    elif suffix and model_dir.endswith(suffix):
        model_file = model_dir
    elif not suffix and alternative_suffix:
        suffix = alternative_suffix
        model_file = model_dir
    else:
        suffix = suffix or default_suffix
        dirobj = mlrun.datastore.store_manager.object(url=model_dir)
        model_dir_list = dirobj.listdir()
        if model_spec_filename in model_dir_list:
            model_spec = _load_model_spec(path.join(model_dir, model_spec_filename))
            model_file = _get_file_path(model_dir, model_spec.model_file, isdir=True)
            extra_dataitems = _get_extra(model_dir, model_spec.extra_data, is_dir=True)
        else:
            extra_dataitems = _get_extra(
                model_dir, {v: v for v in model_dir_list}, is_dir=True
            )
            for file in model_dir_list:
                if file.endswith(suffix):
                    model_file = path.join(model_dir, file)
                    break
    if not model_file:
        raise ValueError(f"cant resolve model file for {model_dir} suffix{suffix}")

    obj = mlrun.datastore.store_manager.object(url=model_file)
    if obj.kind == "file":
        return model_file, model_spec, extra_dataitems

    temp_path = tempfile.NamedTemporaryFile(suffix=suffix, delete=False).name
    obj.download(temp_path)
    return temp_path, model_spec, extra_dataitems


def update_model(
    model_artifact,
    parameters: Optional[dict] = None,
    metrics: Optional[dict] = None,
    extra_data: Optional[dict] = None,
    inputs: Optional[list[Feature]] = None,
    outputs: Optional[list[Feature]] = None,
    feature_vector: Optional[str] = None,
    feature_weights: Optional[list] = None,
    key_prefix: str = "",
    labels: Optional[dict] = None,
    write_spec_copy=True,
    store_object: bool = True,
) -> ModelArtifact:
    """Update model object attributes

    this method will edit or add attributes to a model object

    example::

        update_model(
            model_path,
            metrics={"speed": 100},
            extra_data={"my_data": b"some text", "file": "s3://mybucket/.."},
        )

    :param model_artifact:  model artifact object or path (store://..) or DataItem
    :param parameters:      parameters dict
    :param metrics:         model metrics e.g. accuracy
    :param extra_data:      extra data items key, value dict
                            (value can be: path string | bytes | artifact)
    :param inputs:          list of input features (feature vector schema)
    :param outputs:         list of output features (output vector schema)
    :param feature_vector:  feature store feature vector uri (store://feature-vectors/<project>/<name>[:tag])
    :param feature_weights: list of feature weights, one per input column
    :param key_prefix:      key prefix to add to metrics and extra data items
    :param labels:          metadata labels
    :param write_spec_copy: write a YAML copy of the spec to the target dir
    :param store_object:    Whether to store the model artifact updated.
    """

    if hasattr(model_artifact, "artifact_url"):
        model_artifact = model_artifact.artifact_url

    if isinstance(model_artifact, ModelArtifact):
        model_spec = model_artifact
    elif mlrun.datastore.is_store_uri(model_artifact):
        model_spec, _ = mlrun.datastore.store_manager.get_store_artifact(model_artifact)
    else:
        raise ValueError("model path must be a model store object/URL/DataItem")

    if not model_spec or model_spec.kind != "model":
        raise ValueError(f"store artifact ({model_artifact}) is not model kind")

    if parameters:
        for key, val in parameters.items():
            model_spec.parameters[key] = val
    if metrics:
        for key, val in metrics.items():
            model_spec.metrics[key_prefix + key] = val
    if labels:
        for key, val in labels.items():
            model_spec.labels[key] = val
    if inputs:
        model_spec.inputs = inputs
    if outputs:
        model_spec.outputs = outputs
    if feature_weights:
        model_spec.feature_weights = feature_weights
    if feature_vector:
        model_spec.feature_vector = feature_vector

    if extra_data:
        for key, item in extra_data.items():
            if hasattr(item, "target_path"):
                extra_data[key] = item.target_path

        upload_extra_data(model_spec, extra_data, prefix=key_prefix, update_spec=True)

    if write_spec_copy:
        spec_path = path.join(model_spec.target_path, model_spec_filename)
        model_spec_yaml = _sanitize_and_serialize_model_spec_yaml(model_spec)
        mlrun.datastore.store_manager.object(url=spec_path).put(model_spec_yaml)

    model_spec.db_key = model_spec.db_key or model_spec.key
    if store_object:
        mlrun.get_run_db().store_artifact(
            model_spec.db_key,
            model_spec.to_dict(),
            tree=model_spec.tree,
            iter=model_spec.iter,
            project=model_spec.project,
        )
    return model_spec


def _get_src_path(model_spec: ModelArtifact, filename: str) -> str:
    return path.join(model_spec.src_path, filename) if model_spec.src_path else filename


def _load_model_spec(spec_path) -> ModelArtifact:
    data = mlrun.datastore.store_manager.object(url=spec_path).get()
    spec = yaml.load(data, Loader=yaml.FullLoader)
    return ModelArtifact.from_dict(spec)


def _get_file_path(base_path: str, name: str, isdir: bool = False) -> str:
    if not is_relative_path(name):
        return name
    if not isdir:
        base_path = path.dirname(base_path)
    return path.join(base_path, name).replace("\\", "/")


def _get_extra(target: str, extra_data: dict, is_dir: bool = False) -> dict:
    extra_dataitems = {}
    for k, v in extra_data.items():
        extra_dataitems[k] = mlrun.datastore.store_manager.object(
            url=_get_file_path(target, v, isdir=is_dir), key=k
        )
    return extra_dataitems


def _sanitize_and_serialize_model_spec_yaml(model: ModelArtifact) -> str:
    model_dict = _sanitize_model_spec(model)
    return _serialize_model_spec_yaml(model_dict)


def _sanitize_model_spec(model: ModelArtifact) -> dict:
    model_dict = model.to_dict()

    # The model spec yaml should not include the tag, as the same model can be used with different tags,
    # and the tag is not part of the model spec but the metadata of the model artifact
    model_dict["metadata"].pop("tag", None)

    # Remove future packaging links
    if model_dict["spec"].get("extra_data"):
        model_dict["spec"]["extra_data"] = {
            key: item
            for key, item in model_dict["spec"]["extra_data"].items()
            if item is not ...
        }
    return model_dict


def _serialize_model_spec_yaml(model_dict: dict) -> str:
    return yaml.safe_dump(model_dict)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import json
import os
import socket

import mlrun.db
from mlrun.errors import err_to_str
from mlrun.execution import MLClientCtx
from mlrun.model import RunTemplate
from mlrun.runtimes.local import get_func_arg
from mlrun.serving.server import v2_serving_init
from mlrun.serving.v1_serving import nuclio_serving_init

from .serving import serving_subkind


def nuclio_init_hook(context, data, kind):
    if kind == "serving":
        nuclio_serving_init(context, data)
    elif kind == serving_subkind:
        v2_serving_init(context, data)
    elif kind in ["mlrun", "jobs"]:
        nuclio_jobs_init(context, data)
    else:
        raise ValueError("unsupported nuclio kind")


def nuclio_jobs_init(context, data):
    setattr(context, "mlrun_handler", nuclio_jobs_handler)
    setattr(context, "globals", data)


def nuclio_jobs_handler(context, event):
    paths = event.path.strip("/").split("/")

    if not paths or paths[0] not in context.globals:
        return context.Response(
            body=f"function name {paths[0]} does not exist",
            content_type="text/plain",
            status_code=400,
        )

    fhandler = context.globals[paths[0]]
    if not inspect.isfunction(fhandler) or paths[0].startswith("_"):
        return context.Response(
            body=f"{paths[0]} is not a public function handler",
            content_type="text/plain",
            status_code=400,
        )

    out = mlrun.db.get_or_set_dburl()
    if out:
        context.logger.info(f"logging run results to: {out}")

    newspec = event.body or {}
    if newspec and not isinstance(newspec, dict):
        newspec = json.loads(newspec)

    ctx = MLClientCtx.from_dict(
        newspec,
        rundb=out,
        autocommit=False,
        log_stream=context.logger,
        host=socket.gethostname(),
        store_run=False,
    )

    # Inject project secrets from env. variables to the context
    secret_list = os.environ.get("MLRUN_PROJECT_SECRETS_LIST")
    if secret_list:
        ctx._secrets_manager.add_source("env", secret_list)

    kwargs = get_func_arg(fhandler, RunTemplate.from_dict(ctx.to_dict()), ctx, True)
    try:
        val = fhandler(**kwargs)
        if val:
            ctx.log_result("return", val)
    except Exception as exc:
        err = err_to_str(exc)
        ctx.set_state(error=err)
    return ctx.to_json()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import xgboost as xgb

from .._ml_common import MLTypes, MLUtils


class XGBoostTypes(MLTypes):
    """
    Typing hints for the XGBoost framework.
    """

    # A union of all XGBoost model base classes:
    ModelType = Union[xgb.XGBModel, xgb.Booster]

    # A type for all the supported dataset types:
    DatasetType = Union[MLTypes.DatasetType, xgb.DMatrix]


class XGBoostUtils(MLUtils):
    """
    Utilities functions for the XGBoost framework.
    """

    pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from typing import Optional

import deepdiff
from pydantic.v1 import BaseModel, Extra, Field

import mlrun.common.types
import mlrun.config
import mlrun.errors
from mlrun.common.schemas.object import ObjectKind, ObjectSpec, ObjectStatus


# Defining a different base class (not ObjectMetadata), as there's no project, and it differs enough to
# justify a new class
class HubObjectMetadata(BaseModel):
    name: str
    description: str = ""
    labels: Optional[dict] = {}
    updated: Optional[datetime]
    created: Optional[datetime]

    class Config:
        extra = Extra.allow


class HubSourceType(mlrun.common.types.StrEnum):
    functions = "functions"
    modules = "modules"


# Sources-related objects
class HubSourceSpec(ObjectSpec):
    path: str  # URL to base directory, should include schema (s3://, etc...)
    channel: str
    credentials: Optional[dict] = {}


class HubSource(BaseModel):
    kind: ObjectKind = Field(ObjectKind.hub_source, const=True)
    metadata: HubObjectMetadata
    spec: HubSourceSpec
    status: Optional[ObjectStatus] = ObjectStatus(state="created")

    def get_full_uri(self, relative_path, object_type):
        return f"{self.spec.path}/{object_type}/{self.spec.channel}/{relative_path}"

    def get_catalog_uri(self, object_type):
        return self.get_full_uri(mlrun.mlconf.hub.catalog_filename, object_type)

    @classmethod
    def generate_default_source(cls):
        if not mlrun.mlconf.hub.default_source.create:
            return None

        now = datetime.now(timezone.utc)
        hub_metadata = HubObjectMetadata(
            name=mlrun.mlconf.hub.default_source.name,
            description=mlrun.mlconf.hub.default_source.description,
            created=now,
            updated=now,
        )
        return cls(
            metadata=hub_metadata,
            spec=HubSourceSpec(
                path=mlrun.mlconf.hub.default_source.url,
                channel=mlrun.mlconf.hub.default_source.channel,
            ),
            status=ObjectStatus(state="created"),
        )

    def diff(self, another_source: "HubSource") -> dict:
        """
        Compare this HubSource with another one.
        Returns a dict of differences (metadata, spec, status).
        """
        exclude_paths = [
            "root['metadata']['updated']",
            "root['metadata']['created']",
        ]
        return deepdiff.DeepDiff(
            self.dict(), another_source.dict(), exclude_paths=exclude_paths
        )


last_source_index = -1


class IndexedHubSource(BaseModel):
    index: int = last_source_index  # Default last. Otherwise, must be > 0
    source: HubSource


# Item-related objects
class HubItemMetadata(HubObjectMetadata):
    source: HubSourceType = HubSourceType.functions
    version: str
    tag: Optional[str]

    def get_relative_path(self) -> str:
        # This is needed since the hub deployment script modifies the paths to use _ instead of -.
        modified_name = self.name.replace("-", "_")
        # Prefer using the tag if exists. Otherwise, use version.
        version = self.tag or self.version
        return f"{modified_name}/{version}/"


class HubItemSpec(ObjectSpec):
    item_uri: str
    assets: dict[str, str] = {}


class HubItem(BaseModel):
    kind: ObjectKind = Field(ObjectKind.hub_item, const=True)
    metadata: HubItemMetadata
    spec: HubItemSpec
    status: ObjectStatus


class HubCatalog(BaseModel):
    kind: ObjectKind = Field(ObjectKind.hub_catalog, const=True)
    channel: str
    catalog: list[HubItem]


class HubModuleType(mlrun.common.types.StrEnum):
    generic = "generic"
    monitoring_app = "monitoring_application"



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.types

from .base import ObjectFormat


class FeatureSetFormat(ObjectFormat, mlrun.common.types.StrEnum):
    full = "full"
    minimal = "minimal"

    @staticmethod
    def format_method(_format: str) -> typing.Optional[typing.Callable]:
        return {
            FeatureSetFormat.full: None,
            FeatureSetFormat.minimal: FeatureSetFormat.filter_obj_method(
                [
                    "metadata.name",
                    "metadata.project",
                    "metadata.tag",
                    "metadata.updated",
                    "metadata.uid",
                    "metadata.labels",
                    "spec.entities",
                    "spec.description",
                    "spec.targets",
                    "spec.engine",  # It's not needed by the UI, but we override it anyway to storey if empty
                    "status.state",
                ]
            ),
        }[_format]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from abc import ABCMeta
from types import MethodType
from typing import Any, Optional, Union

import docstring_parser

from mlrun.artifacts import Artifact
from mlrun.datastore import DataItem
from mlrun.utils import logger

from ..errors import MLRunPackagePackingError, MLRunPackageUnpackingError
from ..packager import Packager
from ..utils import DEFAULT_PICKLE_MODULE, ArtifactType, Pickler, TypeHintUtils


class _DefaultPackagerMeta(ABCMeta):
    """
    Metaclass for `DefaultPackager` to override `__doc__` attribute into a class property. This way sphinx will get a
    dynamically generated docstring that will include a summary of the packager.
    """

    def __new__(cls, name: str, bases: tuple, namespace: dict, **kwargs):
        """
        Create a new DefaultPackager metaclass that saves the original packager docstring to another attribute named
        `_packager_doc`.

        :param name:      A string representing the name of the class being instantiated.
        :param bases:     A tuple of classes from which the class will inherit.
        :param namespace: The namespace of the class holding its attributes (from here the docstring will be taken).
        """
        # Save the original doc string to a separate class variable as it will be overriden later on by the metaclass
        # property `__doc__`:
        namespace["_packager_doc"] = namespace.get("__doc__", "")

        # Continue creating the metaclass:
        return super().__new__(cls, name, bases, namespace, **kwargs)

    @property
    def __doc__(cls: type["DefaultPackager"]) -> str:
        """
        Override the `__doc__` attribute of a `DefaultPackager` to be a property in order to auto-summarize the
        packager's class docstring. The summary is concatenated after the original class doc string.

        The summary will be in the following structure:

        <cls._packager_doc>

        .. rubric:: Packager Summary

        **Packing Type**: ``<cls.PACKABLE_OBJECT_TYPE>``

        **Packing Sub-Classes**: True / False

        * **Priority**: ...

        **Default Artifact Types**:

          * **Packing**: ...
          * **Unpacking**: ...

        **Artifact Types**:

        * **type 1**: ...

          * configuration 1 - ...
          * configuration 2 - ...

        * **type 2**: ...

          * configuration 1: ...
          * configuration 2: ...

        :returns: The original docstring with the generated packager summary.
        """
        # Create a packager instance:
        packager = cls()

        # Get the packager's name and module:
        packager_name = packager.__class__.__name__
        packager_module = packager.__module__

        # Get the original packager class doc string:
        packager_doc_string = cls._packager_doc.split("\n")
        packager_doc_string = "\n".join(line[4:] for line in packager_doc_string)

        # Parse the packable type section:
        type_name = (
            "Any type"
            if packager.PACKABLE_OBJECT_TYPE is ...
            else (
                f"``{str(packager.PACKABLE_OBJECT_TYPE)}``"
                if TypeHintUtils.is_typing_type(type_hint=packager.PACKABLE_OBJECT_TYPE)
                else f"``{packager.PACKABLE_OBJECT_TYPE.__module__}.{packager.PACKABLE_OBJECT_TYPE.__name__}``"
            )
        )
        packing_type = f"**Packing Type**: {type_name}"

        # Subclasses support section:
        packing_sub_classes = f"**Packing Sub-Classes**: {packager.PACK_SUBCLASSES}"

        # Priority section:
        priority_value = (
            packager.priority
            if packager.priority is not ...
            else "Default priority (5)"
        )
        priority = f"**Priority**: {priority_value}"

        # Default artifact types:
        def get_default_artifact_type(pack_or_unpack: str) -> str:
            pack_or_unpack = f"default_{pack_or_unpack}ing_artifact_type"
            method_name = f"get_{pack_or_unpack}"
            argument_name = pack_or_unpack.upper()
            return (
                getattr(packager, argument_name)
                if packager_name == "DefaultPackager"
                or method_name not in packager.__class__.__dict__
                else (
                    f"Refer to the packager's "
                    f":py:meth:`~{packager_module}.{packager_name}.{method_name}` method."
                )
            )

        default_artifact_types = (
            f"**Default Artifact Types**:"
            f"\n\n* **Packing**: {get_default_artifact_type('pack')}"
            f"\n\n* **Unpacking**: {get_default_artifact_type('unpack')}"
        )

        # Artifact types section:
        artifact_types = "**Artifact Types**:"
        for artifact_type in packager.get_supported_artifact_types():
            # Get the packing method docstring:
            method_doc = docstring_parser.parse(
                getattr(packager, f"pack_{artifact_type}").__doc__
            )
            # Add the artifact type bullet:
            artifact_type_doc = f"{method_doc.short_description or ''}{method_doc.long_description or ''}".replace(
                "\n", ""
            )
            artifact_types += (
                f"\n\n* :py:meth:`{artifact_type}<{packager_module}.{packager_name}.pack_{artifact_type}>` - "
                + artifact_type_doc
            )
            # Add the artifact type configurations (ignoring the `obj` and `key` parameters):
            configurations_doc = "\n\n  * ".join(
                "{} - {}".format(
                    parameter.arg_name, parameter.description.replace("\n", "")
                )
                for parameter in method_doc.params[2:]
            )
            if configurations_doc:
                artifact_types += f"\n\n  * {configurations_doc}"

        # Construct the final doc string and return:
        doc = (
            f"{packager_doc_string}"
            "\n\n.. rubric:: Packager Summary"
            f"\n\n{packing_type}"
            f"\n\n{packing_sub_classes}"
            f"\n\n{priority}"
            f"\n\n{default_artifact_types}"
            f"\n\n{artifact_types}"
            f"\n\n"
        )
        return doc


class DefaultPackager(Packager, metaclass=_DefaultPackagerMeta):
    """
    A default packager that handles all types and packs them as pickle files.

    The default packager implements all the required methods and has a default logic that should satisfy most
    use cases. To work with this class, don't override the abstract class methods, but instead follow the
    guidelines below:

    * **The class variable** :py:meth:`PACKABLE_OBJECT_TYPE<PACKABLE_OBJECT_TYPE>`: The type of object this packager can
      pack and unpack (used in the ``is_packable`` method).
    * **The class variable** :py:meth:`PACK_SUBCLASSES<PACK_SUBCLASSES>`: A flag that indicates whether to pack all
      subclasses of the ``PACKABLE_OBJECT_TYPE`` (used in the ``is_packable`` method). Default is False.
    * **The class variable** :py:meth:`DEFAULT_PACKING_ARTIFACT_TYPE<DEFAULT_PACKING_ARTIFACT_TYPE>`: The default
      artifact type to pack as. It is returned from the method ``get_default_packing_artifact_type``.
    * **The class variable** :py:meth:`DEFAULT_UNPACKING_ARTIFACT_TYPE<DEFAULT_UNPACKING_ARTIFACT_TYPE>`: The default
      artifact type to unpack from. It is returned from the method
      ``get_default_unpacking_artifact_type``.
    * **The abstract class method** :py:meth:`pack`: This method is implemented to get the object and send it to the
      relevant packing method by the given artifact type using the following naming: "pack_<artifact_type>". (If
      the artifact type was not provided, it uses the default). For example: if the artifact type is `x` then
      the class method ``pack_x`` must be implemented. The signature of each pack class method must be::

          def pack_x(self, obj: Any, key: str, ...) -> Union[Tuple[Artifact, dict], dict]:
              pass

      Where 'x' is the artifact type, 'obj' is the object to pack, `key` is the key to name the artifact and `...` are
      additional, custom, log hint configurations. The returned values are the packed artifact and the instructions
      for unpacking it, or in the case of result, the dictionary of the result with its key and value. configurations
      are sent by the user and shouldn't be mandatory, meaning they should have a default value (otherwise, the user
      has to add them to every log hint).
    * **The abstract class method** :py:meth:`unpack`: The method is implemented to get a
      :py:meth:`DataItem<mlrun.datastore.base.DataItem>` and send it to the relevant unpacking method by the artifact
      type using the following naming: `"unpack_<artifact_type>"`. (If the artifact type was not provided,
      it uses the default).
      For example: if the artifact type stored within the ``DataItem`` is `x` then the class method
      ``unpack_x`` must be implemented. The signature of each unpack class method must be::

          def unpack_x(self, data_item: mlrun.DataItem, ...) -> Any:
              pass

      Where 'x' is the artifact type, 'data_item' is the artifact's data item to unpack, `...` are the instructions that
      were originally returned from ``pack_x``. (Each instruction must be optional (have a default value) to support
      objects from this type that were not packaged but custom-logged.) The returned value is the unpacked
      object.
    * **The abstract class method** :py:meth:`is_packable`: The method is implemented to automatically validate
        the object type and artifact type by the following rules:

      * **Object type validation**: Checks if the given object type matches the variable ``PACKABLE_OBJECT_TYPE``
        with respect to the ``PACK_SUBCLASSES`` class variable.
      * **Artifact type validation**: Checks if the given artifact type is in the list returned from
        ``get_supported_artifact_types``.

    * **The abstract class method** :py:meth:`is_unpackable`: The method is left as implemented in ``Packager``.
    * **The abstract class method** :py:meth:`get_supported_artifact_types`: The method is implemented to look for all
      pack + unpack class methods implemented to collect the supported artifact types. If ``PackagerX`` has ``pack_y``,
      ``unpack_y`` and ``pack_z``, ``unpack_z`` that means the artifact types supported are `y` and `z`.
    * **The abstract class method** :py:meth:`get_default_packing_artifact_type`: The method is implemented to return
      the new class variable ``DEFAULT_PACKING_ARTIFACT_TYPE``. You can still override the method if the default
      artifact type you need could change according to the object that's about to be packed.
    * **The abstract class method** :py:meth:`get_default_unpacking_artifact_type`: The method is implemented to return
      the new class variable ``DEFAULT_UNPACKING_ARTIFACT_TYPE``. You can still override the method if the default
      artifact type you need could change according to the data item that's about to be unpacked.

    .. rubric:: Important to remember

    From the :py:meth:`Packager<mlrun.package.packager.Packager>` docstring:

    * **Linking artifacts** ("extra data"): In order to link between packages (using the extra data or metrics spec
      attributes of an artifact), use the key as if it exists and as value ellipses (...). The manager
      links all packages once it is done packing.

      For example, given extra data keys in the log hint as `extra_data`, set them to an artifact as follows::

          artifact = Artifact(key="my_artifact")
          artifact.spec.extra_data = {key: ... for key in extra_data}

    * **Clearing outputs**: Some packagers may produce files and temporary directories that should be deleted after
      the artifact is logged. The packager can mark paths of files and directories to delete after
      logging using the class method ``add_future_clearing_path``.

      For example, in the following packager's ``pack`` method, you can write a text file, create an artifact, and then
      mark the text file to be deleted once the artifact is logged::

          with open("./some_file.txt", "w") as file:
              file.write("Pack me")
          artifact = Artifact(key="my_artifact")
          self.add_future_clearing_path(path="./some_file.txt")
          return artifact, None

    """

    #: The type of object this packager can pack and unpack.
    PACKABLE_OBJECT_TYPE: type = ...

    #: A flag for indicating whether to also pack all subclasses of the `PACKABLE_OBJECT_TYPE`.
    PACK_SUBCLASSES = False

    #: The default artifact type to pack as.
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.OBJECT

    #: The default artifact type to unpack from.
    DEFAULT_UNPACKING_ARTIFACT_TYPE = ArtifactType.OBJECT

    def get_default_packing_artifact_type(self, obj: Any) -> str:
        """
        Get the default artifact type for packing an object of this packager.

        :param obj: The about-to-be packed object.

        :return: The default artifact type.
        """
        return self.DEFAULT_PACKING_ARTIFACT_TYPE

    def get_default_unpacking_artifact_type(self, data_item: DataItem) -> str:
        """
        Get the default artifact type used for unpacking a data item holding an object of this packager. The method
        is used when a data item is sent for unpacking without it being a package, but is a simple url or an old /
        manually logged artifact.

        :param data_item: The about-to-be unpacked data item.

        :return: The default artifact type.
        """
        return self.DEFAULT_UNPACKING_ARTIFACT_TYPE

    def get_supported_artifact_types(self) -> list[str]:
        """
        Get all the supported artifact types on this packager.

        :return: A list of all the supported artifact types.
        """
        # We look for pack + unpack method couples so there won't be a scenario where an object can be packed but not
        # unpacked. Result has no unpacking so we add it separately.
        return [
            key[len("pack_") :]
            for key in dir(self)
            if key.startswith("pack_") and f"unpack_{key[len('pack_'):]}" in dir(self)
        ] + ["result"]

    def pack(
        self,
        obj: Any,
        key: Optional[str] = None,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> Union[tuple[Artifact, dict], dict]:
        """
        Pack an object as the given artifact type using the provided configurations.

        :param obj:            The object to pack.
        :param key:            The key of the artifact.
        :param artifact_type:  Artifact type to log to MLRun. If passing `None`, the default artifact type is used.
        :param configurations: Log hints configurations to pass to the packing method.

        :return: If the packed object is an artifact, a tuple of the packed artifact and unpacking instructions
                 dictionary. If the packed object is a result, a dictionary containing the result key and value.
        """
        # Get default artifact type in case it was not provided:
        if artifact_type is None:
            artifact_type = self.get_default_packing_artifact_type(obj=obj)

        # Set empty dictionary in case no configurations were given:
        configurations = configurations or {}

        # Get the packing method according to the artifact type:
        pack_method = getattr(self, f"pack_{artifact_type}")

        # Validate correct configurations were passed:
        self._validate_method_arguments(
            method=pack_method,
            arguments=configurations,
            is_packing=True,
        )

        # Call the packing method and return the package:
        return pack_method(obj=obj, key=key, **configurations)

    def unpack(
        self,
        data_item: DataItem,
        artifact_type: Optional[str] = None,
        instructions: Optional[dict] = None,
    ) -> Any:
        """
        Unpack the data item's artifact by the provided type using the given instructions.

        :param data_item:     The data input to unpack.
        :param artifact_type: The artifact type to unpack the data item as. If passing `None`, the default artifact type
                              is used.
        :param instructions:  Additional instructions noted in the package to pass to the unpacking method.

        :return: The unpacked data item's object.

        :raise MLRunPackageUnpackingError: In case the packager could not unpack the data item.
        """
        # Get default artifact type in case it was not provided:
        if artifact_type is None:
            artifact_type = self.get_default_unpacking_artifact_type(
                data_item=data_item
            )

        # Set empty dictionary in case no instructions were given:
        instructions = instructions or {}

        # Get the unpacking method according to the artifact type:
        unpack_method = getattr(self, f"unpack_{artifact_type}")

        # Validate correct instructions were passed:
        self._validate_method_arguments(
            method=unpack_method,
            arguments=instructions,
            is_packing=False,
        )

        # Call the unpacking method and return the object:
        return unpack_method(data_item, **instructions)

    def is_packable(
        self,
        obj: Any,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> bool:
        """
        Check if this packager can pack an object of the provided type as the provided artifact type.

        The method is implemented to validate the object's type and artifact type by checking if the given object type
        matches the variable ``PACKABLE_OBJECT_TYPE`` with respect to the ``PACK_SUBCLASSES`` class variable. If it
        does, it checks if the given artifact type is in the list returned from ``get_supported_artifact_types``.

        :param obj:            The object to pack.
        :param artifact_type:  The artifact type to log the object as.
        :param configurations: The log hint configurations passed by the user.

        :return: True if packable and False otherwise.
        """
        # Get the object's type:
        object_type = type(obj)

        # Check type (ellipses means any type):
        if self.PACKABLE_OBJECT_TYPE is not ...:
            if not TypeHintUtils.is_matching(
                object_type=object_type,
                type_hint=self.PACKABLE_OBJECT_TYPE,
                include_subclasses=self.PACK_SUBCLASSES,
                reduce_type_hint=False,
            ):
                return False

        # Check the artifact type:
        if (
            artifact_type is not None
            and artifact_type not in self.get_supported_artifact_types()
        ):
            return False

        # Packable:
        return True

    def pack_object(
        self,
        obj: Any,
        key: str,
        pickle_module_name: str = DEFAULT_PICKLE_MODULE,
    ) -> tuple[Artifact, dict]:
        """
        Pack a python object, pickling it into a pkl file and store it in an artifact.

        :param obj:                The object to pack and log.
        :param key:                The artifact's key.
        :param pickle_module_name: The pickle module name to use for serializing the object.

        :return: The artifacts and its pickling instructions.
        """
        # Pickle the object to file:
        pickle_path, instructions = Pickler.pickle(
            obj=obj, pickle_module_name=pickle_module_name
        )

        # Initialize an artifact to the pkl file:
        artifact = Artifact(key=key, src_path=pickle_path)

        # Add the pickle path to the clearing list:
        self.add_future_clearing_path(path=pickle_path)

        return artifact, instructions

    def pack_result(self, obj: Any, key: str) -> dict:
        """
        Pack an object as a result.

        :param obj: The object to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return {key: obj}

    def unpack_object(
        self,
        data_item: DataItem,
        pickle_module_name: str = DEFAULT_PICKLE_MODULE,
        object_module_name: Optional[str] = None,
        python_version: Optional[str] = None,
        pickle_module_version: Optional[str] = None,
        object_module_version: Optional[str] = None,
    ) -> Any:
        """
        Unpack the data item's object, unpickle it using the instructions, and return.

        Warnings of mismatching python and module versions between the original pickling interpreter and this one may be
        raised.

        :param data_item:             The data item holding the pkl file.
        :param pickle_module_name:    Module to use for unpickling the object.
        :param object_module_name:    The original object's module. Used to verify that the current interpreter object
                                      module version matches the pickled object version before unpickling the object.
        :param python_version:        The python version in which the original object was pickled. Used to verify that
                                      the current interpreter python version matches the pickled object version before
                                      unpickling the object.
        :param pickle_module_version: The pickle module version. Used to verify that the current interpreter module
                                      version matches the one that pickled the object before unpickling it.
        :param object_module_version: The original object's module version to match to the interpreter's module version.

        :return: The un-pickled python object.
        """
        # Get the pkl file to local directory:
        pickle_path = self.get_data_item_local_path(data_item=data_item)

        # Unpickle and return:
        return Pickler.unpickle(
            pickle_path=pickle_path,
            pickle_module_name=pickle_module_name,
            object_module_name=object_module_name,
            python_version=python_version,
            pickle_module_version=pickle_module_version,
            object_module_version=object_module_version,
        )

    def _validate_method_arguments(
        self, method: MethodType, arguments: dict, is_packing: bool
    ):
        """
        Validate keyword arguments to pass to a method. Used for validating log hint configurations for packing methods
        and instructions for unpacking methods.

        :param method:     The method to validate the arguments for.
        :param arguments:  Keyword arguments to validate.
        :param is_packing: Flag to know if the arguments came from packing or unpacking, to raise the correct exception
                           if validation failed.

        :raise MLRunPackagePackingError:   If there are missing configurations in the log hint.
        :raise MLRunPackageUnpackingError: If there are missing instructions in the artifact's spec.
        """
        # Get the possible and mandatory (arguments that has no default value) arguments from the functions:
        possible_arguments = inspect.signature(method).parameters
        mandatory_arguments = [
            name
            for name, parameter in possible_arguments.items()
            # If default value is `empty` it is mandatory:
            if parameter.default is inspect.Parameter.empty
            # Ignore the *args and **kwargs parameters:
            and parameter.kind
            not in [inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL]
        ]

        # Remove non-configuration (or non-instructions in case of unpacking) arguments:
        if is_packing:
            mandatory_arguments.remove("obj")
            if "key" in mandatory_arguments:
                # Default keys may be an option in the future, so it is not necessary a mandatory argument:
                mandatory_arguments.remove("key")
        else:
            mandatory_arguments.remove("data_item")

        # Validate there are no missing arguments (only mandatory ones):
        missing_arguments = [
            mandatory_argument
            for mandatory_argument in mandatory_arguments
            if mandatory_argument not in arguments
        ]
        if missing_arguments:
            if is_packing:
                raise MLRunPackagePackingError(
                    f"The packager '{self.__class__.__name__}' could not pack the package due to missing "
                    f"configurations: {', '.join(missing_arguments)}. Add the missing arguments to the log hint of "
                    f"this object in order to pack it. Make sure you pass a dictionary log hint and not a string in "
                    f"order to pass configurations in the log hint."
                )
            raise MLRunPackageUnpackingError(
                f"The packager '{self.__class__.__name__}' could not unpack the package due to missing instructions: "
                f"{', '.join(missing_arguments)}. Missing instructions are likely due to an update in the packager's "
                f"code that not support the old implementation. This backward compatibility should not occur. To "
                f"overcome it, try to edit the instructions in the artifact's spec to enable unpacking it again."
            )

        # Validate all given arguments are correct:
        incorrect_arguments = [
            argument for argument in arguments if argument not in possible_arguments
        ]
        if incorrect_arguments:
            arguments_type = "configurations" if is_packing else "instructions"
            logger.warn(
                f"Unexpected {arguments_type} given for {self.__class__.__name__}: {', '.join(incorrect_arguments)}. "
                f"Possible {arguments_type} are: {', '.join(possible_arguments.keys())}. The packager tries to "
                f"continue by ignoring the incorrect arguments."
            )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
import typing
import uuid
from os import environ, path

import pandas as pd

import mlrun.utils

from .config import config
from .datastore import uri_to_ipython
from .utils import dict_to_list, get_in, is_jupyter

JUPYTER_SERVER_ROOT = environ.get("HOME", "/User")
supported_viewers = [
    ".htm",
    ".html",
    ".json",
    ".yaml",
    ".txt",
    ".log",
    ".jpg",
    ".png",
    ".csv",
    ".py",
]


def html_dict(title, data, open=False, show_nil=False):
    if not data:
        return ""
    html = ""
    for key, val in data.items():
        if show_nil or val:
            html += f"<tr><th>{key}</th><td>{val}</td></tr>"
    if html:
        html = f"<table>{html}</table>"
        return html_summary(title, html, open=open)
    return ""


def html_summary(title, data, num=None, open=False):
    tag = ""
    if open:
        tag = " open"
    if num:
        title = f"{title} ({num})"
    return f"<details{tag}><summary><b>{title}<b></summary>{data}</details>"


def html_crop(x):
    return f'<div class="ellipsis" ondblclick="copyToClipboard(this)" title="{x} (dbl click to copy)">{x}</div>'


def table_sum(title, df):
    size = len(df.index)
    if size > 0:
        return html_summary(title, df.to_html(escape=False), size)


def dict_html(x):
    return "".join([f'<div class="dictlist">{i}</div>' for i in dict_to_list(x)])


def link_to_ipython(link: str):
    """
    Convert a link (e.g. v3io path) to a jupyter notebook local link.

    :param link: the link to convert
    :return:     the converted link and ref for expanding the file in the notebook
    """
    valid = pathlib.Path(link).suffix in supported_viewers
    ref = 'class="artifact" onclick="expandPanel(this)" paneName="result" '
    if "://" not in link:
        abs = path.abspath(link)
        if abs.startswith(JUPYTER_SERVER_ROOT) and valid:
            return abs.replace(JUPYTER_SERVER_ROOT, "/files"), ref
        else:
            return abs, ""
    else:
        newlink = uri_to_ipython(link)
        if newlink and valid:
            return "files/" + newlink, ref
    return link, ""


def link_html(text, link=""):
    if not link:
        link = text
    link, ref = link_to_ipython(link)
    return f'<div {ref}title="{link}">{text}</div>'


def artifacts_html(
    artifacts: list[dict],
    attribute_name: str = "path",
):
    """
    Generate HTML for a list of artifacts. The HTML will be a list of links to the artifacts to be presented in the
    jupyter notebook. The links will be clickable and will open the artifact in a new tab.

    :param artifacts:       contains a list of artifact dictionaries
    :param attribute_name:  the attribute of the artifact to use as the link text
    :return:                the generated HTML
    """
    if not artifacts:
        return ""
    html = ""

    for artifact in artifacts:
        attribute_value = artifact["spec"].get(attribute_name)
        key = artifact["metadata"]["key"]

        if not attribute_value:
            mlrun.utils.logger.warning(
                f"Artifact required attribute {attribute_name} is missing, omitting from output",
                artifact_key=key,
            )
            continue

        link, ref = link_to_ipython(attribute_value)
        html += f'<div {ref}title="{link}">{key}</div>'
    return html


def inputs_html(x):
    if not x:
        return ""
    html = ""
    for k, v in x.items():
        link, ref = link_to_ipython(v)
        html += f'<div {ref}title="{link}">{k}</div>'
    return html


def sources_list_html(x):
    if not x:
        return ""
    html = ""
    for src in x:
        v = src.get("path", "")
        link, ref = link_to_ipython(v)
        html += f'<div {ref}title="{link}">{src["name"]}</div>'
    return html


def run_to_html(results, display=True):
    html = html_dict("Metadata", results["metadata"])
    html += html_dict("Spec", results["spec"])
    html += html_dict("results", results["status"].get("results"), True, True)

    if "iterations" in results["status"]:
        iter = results["status"]["iterations"]
        if iter:
            df = pd.DataFrame(iter[1:], columns=iter[0]).set_index("iter")
            html += table_sum("Iterations", df)

    artifacts = results["status"].get("artifacts", None)
    if artifacts:
        df = pd.DataFrame(artifacts)
        if "description" not in df.columns.values:
            df["description"] = ""
        df = df[["key", "kind", "target_path", "description"]]
        df["target_path"] = df["target_path"].apply(link_html)
        html += table_sum("Artifacts", df)

    return ipython_display(html, display)


def ipython_display(html, display=True, alt_text=None):
    if display and html and is_jupyter:
        import IPython.display

        IPython.display.display(IPython.display.HTML(html))
    elif alt_text:
        print(alt_text)
    return html


def get_style():
    return f"""<style>
.dictlist {{
  background-color: {config.background_color};
  text-align: center;
  margin: 4px;
  border-radius: 3px; padding: 0px 3px 1px 3px; display: inline-block;}}
.artifact {{
  cursor: pointer;
  background-color: {config.background_color};
  text-align: left;
  margin: 4px; border-radius: 3px; padding: 0px 3px 1px 3px; display: inline-block;
}}
div.block.hidden {{
  display: none;
}}
.clickable {{
  cursor: pointer;
}}
.ellipsis {{
  display: inline-block;
  max-width: 60px;
  white-space: nowrap;
  overflow: hidden;
  text-overflow: ellipsis;
}}
.master-wrapper {{
  display: flex;
  flex-flow: row nowrap;
  justify-content: flex-start;
  align-items: stretch;
}}
.master-tbl {{
  flex: 3
}}
.master-wrapper > div {{
  margin: 4px;
  padding: 10px;
}}
iframe.fileview {{
  border: 0 none;
  height: 100%;
  width: 100%;
  white-space: pre-wrap;
}}
.pane-header-title {{
  width: 80%;
  font-weight: 500;
}}
.pane-header {{
  line-height: 1;
  background-color: {config.background_color};
  padding: 3px;
}}
.pane-header .close {{
  font-size: 20px;
  font-weight: 700;
  float: right;
  margin-top: -5px;
}}
.master-wrapper .right-pane {{
  border: 1px inset silver;
  width: 40%;
  min-height: 300px;
  flex: 3
  min-width: 500px;
}}
.master-wrapper * {{
  box-sizing: border-box;
}}
</style>"""


jscripts = r"""<script>
function copyToClipboard(fld) {
    if (document.queryCommandSupported && document.queryCommandSupported('copy')) {
        var textarea = document.createElement('textarea');
        textarea.textContent = fld.innerHTML;
        textarea.style.position = 'fixed';
        document.body.appendChild(textarea);
        textarea.select();

        try {
            return document.execCommand('copy'); // Security exception may be thrown by some browsers.
        } catch (ex) {

        } finally {
            document.body.removeChild(textarea);
        }
    }
}
function expandPanel(el) {
  const panelName = "#" + el.getAttribute('paneName');

  // Get the base URL of the current notebook
  var baseUrl = window.location.origin;

  // Construct the full URL
  var fullUrl = new URL(el.title, baseUrl).href;

  document.querySelector(panelName + "-title").innerHTML = fullUrl
  iframe = document.querySelector(panelName + "-body");

  const tblcss = `<style> body { font-family: Arial, Helvetica, sans-serif;}
    #csv { margin-bottom: 15px; }
    #csv table { border-collapse: collapse;}
    #csv table td { padding: 4px 8px; border: 1px solid silver;} </style>`;

  function csvToHtmlTable(str) {
    return '<div id="csv"><table><tr><td>' +  str.replace(/[\n\r]+$/g, '').replace(/[\n\r]+/g, '</td></tr><tr><td>')
      .replace(/,/g, '</td><td>') + '</td></tr></table></div>';
  }

  function reqListener () {
    if (fullUrl.endsWith(".csv")) {
      iframe.setAttribute("srcdoc", tblcss + csvToHtmlTable(this.responseText));
    } else {
      iframe.setAttribute("srcdoc", this.responseText);
    }
    console.log(this.responseText);
  }

  const oReq = new XMLHttpRequest();
  oReq.addEventListener("load", reqListener);
  oReq.open("GET", fullUrl);
  oReq.send();


  //iframe.src = fullUrl;
  const resultPane = document.querySelector(panelName + "-pane");
  if (resultPane.classList.contains("hidden")) {
    resultPane.classList.remove("hidden");
  }
}
function closePanel(el) {
  const panelName = "#" + el.getAttribute('paneName')
  const resultPane = document.querySelector(panelName + "-pane");
  if (!resultPane.classList.contains("hidden")) {
    resultPane.classList.add("hidden");
  }
}

</script>"""

tblframe = """
<div class="master-wrapper">
  <div class="block master-tbl">{}</div>
  <div id="result-pane" class="right-pane block hidden">
    <div class="pane-header">
      <span id="result-title" class="pane-header-title">Title</span>
      <span onclick="closePanel(this)" paneName="result" class="close clickable">&times;</span>
    </div>
    <iframe class="fileview" id="result-body"></iframe>
  </div>
</div>
"""


def get_tblframe(df, display, classes=None):
    table_html = df.to_html(
        escape=False, index=False, notebook=display, classes=classes
    )
    if not display:
        return table_html

    table = tblframe.format(table_html)
    rnd = "result" + str(uuid.uuid4())[:8]
    html = get_style() + jscripts + table.replace('="result', '="' + rnd)
    return ipython_display(html, display)


def runs_to_html(
    df: pd.DataFrame,
    display: bool = True,
    classes: typing.Optional[typing.Union[str, list, tuple]] = None,
    short: bool = False,
):
    def time_str(x):
        try:
            return x.strftime("%b %d %H:%M:%S")
        except ValueError:
            return ""

    df["results"] = df["results"].apply(dict_html)
    df["start"] = df["start"].apply(time_str)
    df["parameters"] = df["parameters"].apply(dict_html)
    uid_template = '<div title="{}"><a href="{}" target="_blank" >...{}</a></div>'

    if config.resolve_ui_url():
        df["uid"] = df.apply(
            lambda x: uid_template.format(
                x["uid"],
                mlrun.utils.get_run_url(x["project"], x["uid"], x["name"]),
                x["uid"][-8:],
            ),
            axis=1,
        )
    else:
        df["uid"] = df["uid"].apply(lambda x: f'<div title="{x}">...{x[-6:]}</div>')

    if short:
        df.drop("project", axis=1, inplace=True)
        if df["iter"].nunique() == 1:
            df.drop("iter", axis=1, inplace=True)
        df.drop("labels", axis=1, inplace=True)
        df.drop("inputs", axis=1, inplace=True)
        df.drop("artifacts", axis=1, inplace=True)
        df.drop("artifact_uris", axis=1, inplace=True)
    else:
        df["labels"] = df["labels"].apply(dict_html)
        df["inputs"] = df["inputs"].apply(inputs_html)
        if df["artifacts"][0]:
            df["artifacts"] = df["artifacts"].apply(
                lambda artifacts: artifacts_html(artifacts, "target_path"),
            )
            df.drop("artifact_uris", axis=1, inplace=True)
        elif df["artifact_uris"][0]:
            df["artifact_uris"] = df["artifact_uris"].apply(dict_html)
            df.drop("artifacts", axis=1, inplace=True)
        else:
            df.drop("artifacts", axis=1, inplace=True)
            df.drop("artifact_uris", axis=1, inplace=True)

    def expand_error(x):
        if x["state"] == "error":
            title = str(x["error"])
            state = f'<div style="color: red;" title="{title}">{x["state"]}</div>'
            x["state"] = state
        return x

    df = df.apply(expand_error, axis=1)
    df.drop("error", axis=1, inplace=True)
    with pd.option_context("display.max_colwidth", None):
        return get_tblframe(df, display, classes=classes)


def artifacts_to_html(df, display=True, classes=None):
    def prod_htm(x):
        if not x or not isinstance(x, dict):
            return ""
        kind = get_in(x, "kind", "")
        uri = get_in(x, "uri", "")
        name = get_in(x, "name", "unknown")
        title = f"{kind}/{uri}"
        if "owner" in x:
            title += f" by {x['owner']}"
        return f'<div title="{title}" class="producer">{name}</div>'

    if "tree" in df.columns.values:
        df["tree"] = df["tree"].apply(html_crop)
    df["path"] = df["path"].apply(link_html)
    df["hash"] = df["hash"].apply(html_crop)
    df["sources"] = df["sources"].apply(sources_list_html)
    df["labels"] = df["labels"].apply(dict_html)
    df["producer"] = df["producer"].apply(prod_htm)
    df["updated"] = df["updated"].apply(lambda x: x.strftime("%b %d %H:%M:%S"))
    with pd.option_context("display.max_colwidth", None):
        return get_tblframe(df, display, classes=classes)
